# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Helpers for the implementation of RFC7959 blockwise transfers"""

import types

from . import numbers
from .numbers.optionnumbers import OptionNumber
from .error import ConstructionRenderableError
from .message import Message
from .optiontypes import BlockOption
from .util.asyncio.timeoutdict import TimeoutDict

def _extract_block_key(message):
    """Extract a key that hashes equally for all blocks of a blockwise
    operation from a request message.

    See discussion at <https://mailarchive.ietf.org/arch/msg/core/I-6LzAL6lIUVDA6_g9YM3Zjhg8E>.
    """

    return (message.remote.blockwise_key, message.code, message.get_cache_key([
        OptionNumber.BLOCK1,
        OptionNumber.BLOCK2,
        OptionNumber.OBSERVE,
        ]))

class ContinueException(ConstructionRenderableError):
    """Not an error in the CoAP sense, but an error in the processing sense,
    indicating that no complete request message is available for processing.

    It reflects back the request's block1 option when rendered.
    """
    def __init__(self, block1):
        self.block1 = block1

    def to_message(self):
        m = super().to_message()
        m.opt.block1 = self.block1
        return m

    code = numbers.CONTINUE

class IncompleteException(ConstructionRenderableError):
    code = numbers.REQUEST_ENTITY_INCOMPLETE

class Block1Spool:
    def __init__(self):
        # FIXME: introduce an actual parameter here
        self._assemblies = TimeoutDict(numbers.TransportTuning().MAX_TRANSMIT_WAIT)

    def feed_and_take(self, req: Message) -> Message:
        """Assemble the request into the spool. This either produces a
        reassembled request message, or raises either a Continue or a Request
        Entity Incomplete exception.

        Requests without block1 are simply passed through."""

        if req.opt.block1 is None:
            return req

        block_key = _extract_block_key(req)

        if req.opt.block1.block_number == 0:
            # silently discarding any old incomplete operation
            self._assemblies[block_key] = req
        else:
            try:
                self._assemblies[block_key]._append_request_block(req)
            except KeyError:
                # KeyError: Received unmatched blockwise response
                # ValueError: Failed to assemble -- gaps or overlaps in data
                raise IncompleteException from None

        if req.opt.block1.more:
            raise ContinueException(req.opt.block1)
        else:
            return self._assemblies[block_key]
            # which happens to carry the last block's block1 option

class Block2Cache:
    """A cache of responses to a give block key.

    Use this when result rendering is expensive, not idempotent or has varying
    output -- otherwise it's often better to calculate the full response again
    and serve chunks.
    """
    def __init__(self):
        # FIXME: introduce an actual parameter here
        self._completes = TimeoutDict(numbers.TransportTuning().MAX_TRANSMIT_WAIT)

    async def extract_or_insert(self, req: Message, response_builder: types.CoroutineType):
        """Given a request message,

        * if it is querying a particular block, look it up in the cache or
          raise Request Entity Incomplete.
        * otherwise,
          * await the response builder
          * return the response if it doesn't need chunking, or
          * return the first chunk and store it for later use

        """
        block_key = _extract_block_key(req)

        if req.opt.block2 is None or req.opt.block2.block_number == 0:
            assembled = await response_builder()
        else:
            try:
                assembled = self._completes[block_key]
            except KeyError:
                raise IncompleteException from None

        if len(assembled.payload) > req.remote.maximum_payload_size or \
                req.opt.block2 is not None and len(assembled.payload) > req.opt.block2.size:
            self._completes[block_key] = assembled

            block2 = req.opt.block2 or \
                    BlockOption.BlockwiseTuple(0, 0, req.remote.maximum_block_size_exp)
            return assembled._extract_block(
                    block2.block_number,
                    block2.size_exponent,
                    req.remote.maximum_payload_size
                    )
        else:
            return assembled



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""
The aiocoap package is a library that implements CoAP, the `Constrained
Application Protocol`_.

If you are reading this through the Python documentation, be aware that there
is additional documentation available online_ and in the source code's ``doc``
directory.

.. _`Constrained Application Protocol`: http://coap.technology/
.. _online: http://aiocoap.readthedocs.io/

Module contents
---------------

This root module re-exports the most commonly used classes in aiocoap:
:class:`.Context`, :class:`.Message` as well as all commonly used numeric
constants from :mod:`.numbers`; see their respective documentation entries.

The presence of :class:`.Message` and :class:`.Context` in the root module is
stable.
"""

import numbers
# flake8 doesn't see through the global re-export
from .numbers import * # noqa: F401, F403
from .message import Message, NoResponse
from .protocol import Context

__all__ = numbers.__all__ + ['Message', 'NoResponse', 'Context']



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Client components for draft-ietf-core-resource-directory

Currently, this only contains the :mod:`register` module, which helps utilizing
the registration interface of a resource directory.
"""



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This module contains all internals needed to manage messages on unreliable
transports, ie. everything that deals in message types or Message IDs.

Currently, it also provides the mechanisms for managing tokens, but those will
be split into dedicated classes.
"""

import asyncio
import functools
import random
from typing import Dict, Tuple, Optional

from . import error
from . import interfaces
from .interfaces import EndpointAddress
from .message import Message
from .numbers.types import CON, ACK, RST, NON
from .numbers.codes import EMPTY


class MessageManager(interfaces.TokenInterface, interfaces.MessageManager):
    """This MessageManager Drives a message interface following the rules of
    RFC7252 CoAP over UDP.

    It takes care of picking message IDs (mid) for outgoing messages,
    retransmitting CON messages, and to react appropriately to incoming
    messages' type, sending ACKs either immediately or later.

    It creates piggy-backed responses by keeping an eye on the tokens the
    messages are sent with, but otherwise ignores the tokens. (It inspects
    tokens *only* where required by its sub-layer).
    """

    def __init__(self, token_manager):
        self.token_manager = token_manager

        self.message_id = random.randint(0, 65535)
        #: Tracker of recently received messages (by remote and message ID).
        #: Maps them to a response message when one is already known.
        self._recent_messages: Dict[Tuple[EndpointAddress, int], Optional[Message]] = {}
        self._active_exchanges = {}  #: active exchanges i.e. sent CON messages (remote, message-id): (messageerror_monitor monitor, cancellable timeout)
        self._backlogs = {} #: per-remote list of (backlogged package, messageerror_monitor) tupless (keys exist iff there is an active_exchange with that node)

        #: Maps pending remote/token combinations to the MID a response can be
        #: piggybacked on, and the timeout that should be cancelled if it is.
        self._piggyback_opportunities: Dict[Tuple[EndpointAddress, bytes], (int, asyncio.TimerHandle)] = {}

        self.log = token_manager.log
        self.loop = token_manager.loop

        #self.message_interface = … -- needs to be set post-construction, because the message_interface in its constructor already needs to get its manager

    def __repr__(self):
        return '<%s for %s>' % (type(self).__name__, getattr(self, 'message_interface', '(unbound)'))

    @property
    def client_credentials(self):
        return self.token_manager.client_credentials

    async def shutdown(self):
        for messageerror_monitor, cancellable in self._active_exchanges.values():
            # Not calling messageerror_monitor: This is not message specific,
            # and its shutdown will take care of these things
            cancellable.cancel()
        self._active_exchanges = None

        await self.message_interface.shutdown()

    #
    # implementing the MessageManager interface
    #

    def dispatch_message(self, message):
        """Feed a message through the message-id, message-type and message-code
        sublayers of CoAP"""

        self.log.debug("Incoming message %r", message)
        if message.code.is_request():
            # Responses don't get deduplication because they "are idempotent or
            # can be handled in an idempotent fashion" (RFC 7252 Section 4.5).
            # This means that a separate response may get a RST when it is
            # arrives at the aiocoap client twice. Note that this does not
            # impede the operation of observations: Their token is still active
            # so they are ACK'd, and deduplication based on observation numbers
            # filters out the rest.
            #
            # This saves memory, and allows stateful transports to be shut down
            # expeditiously unless kept alive by something else (otherwise,
            # they'd linger for EXCHANGE_LIFETIME with no good reason).
            if self._deduplicate_message(message) is True:
                return

        if message.mtype in (ACK, RST):
            self._remove_exchange(message)

        if message.code is EMPTY and message.mtype is CON:
            self._process_ping(message)
        elif message.code is EMPTY and message.mtype in (ACK, RST):
            pass # empty ack has already been handled above
        elif message.code.is_request() and message.mtype in (CON, NON):
            # the request handler will have to deal with sending ACK itself, as
            # it might be timeout-related
            self._process_request(message)
        elif message.code.is_response() and message.mtype in (CON, NON, ACK):
            success = self._process_response(message)
            if success:
                if message.mtype is CON:
                    self._send_empty_ack(message.remote, message.mid, reason="acknowledging incoming response")
            else:
                # A peer mustn't send a CON to multicast, but if a malicious
                # peer does, we better not answer
                if message.mtype == CON and not message.remote.is_multicast_locally:
                    self.log.info("Response not recognized - sending RST.")
                    rst = Message(mtype=RST, mid=message.mid, code=EMPTY, payload='')
                    rst.remote = message.remote.as_response_address()
                    self._send_initially(rst)
                else:
                    self.log.info("Ignoring unknown response (which is not a unicast CON)")
        else:
            self.log.warning("Received a message with code %s and type %s (those don't fit) from %s, ignoring it.", message.code, message.mtype, message.remote)

    def dispatch_error(self, error, remote):
        if self._active_exchanges is None:
            # Not entirely sure where it is so far; better just raise a warning
            # than an exception later, nothing terminally bad should come of
            # this error.
            self.log.warning("Internal shutdown sequence msismatch: error dispatched through messagemanager after shutown")
            return

        self.log.debug("Incoming error %s from %r", error, remote)

        # cancel requests first, and then exchanges: cancelling the pending
        # exchange would trigger enqueued requests to be transmitted
        self.token_manager.dispatch_error(error, remote)

        keys_for_removal = []
        for key, (messageerror_monitor, cancellable_timeout) in self._active_exchanges.items():
            (exchange_remote, message_id) = key
            if remote == exchange_remote:
                keys_for_removal.append(key)
        for k in keys_for_removal:
            (messageerror_monitor, cancellable_timeout) = self._active_exchanges.pop(k)
            cancellable_timeout.cancel()
            # not triggering the messageerror_monitor: that already got the
            # clue from the token manager
        self._backlogs.pop(remote, ())
        # while that's an iterable over messages and messageerror monitors, not
        # triggering them either for th esame reason as above

    #
    # coap dispatch, message-id sublayer: duplicate handling
    #

    def _deduplicate_message(self, message):
        """Return True if a message is a duplicate, and re-send the stored
        response if available.

        Duplicate is a message with the same Message ID (mid) and sender
        (remote), as message received within last EXCHANGE_LIFETIME seconds
        (usually 247 seconds)."""

        key = (message.remote, message.mid)
        if key in self._recent_messages:
            if message.mtype is CON:
                if self._recent_messages[key] is not None:
                    self.log.info('Duplicate CON received, sending old response again')
                    # not going via send_message because that would strip the
                    # mid and might do all other sorts of checks
                    self._send_initially(self._recent_messages[key])
                else:
                    self.log.info('Duplicate CON received, no response to send yet')
            else:
                self.log.info('Duplicate NON, ACK or RST received')
            return True
        else:
            self.log.debug('New unique message received')
            self.loop.call_later(message.transport_tuning.EXCHANGE_LIFETIME, functools.partial(self._recent_messages.pop, key))
            self._recent_messages[key] = None
            return False

    def _store_response_for_duplicates(self, message):
        """If the message is the response can be used to satisfy a future
        duplicate message, store it."""

        key = (message.remote, message.mid)
        if key in self._recent_messages:
            self._recent_messages[key] = message

    #
    # coap dispatch, message-type sublayer: retransmission handling
    #

    def _add_exchange(self, message, messageerror_monitor):
        """Add an "exchange" for outgoing CON message.

        CON (Confirmable) messages are automatically retransmitted by protocol
        until ACK or RST message with the same Message ID is received from
        target host."""

        key = (message.remote, message.mid)

        if message.remote not in self._backlogs:
            self._backlogs[message.remote] = []

        timeout = random.uniform(message.transport_tuning.ACK_TIMEOUT, message.transport_tuning.ACK_TIMEOUT * message.transport_tuning.ACK_RANDOM_FACTOR)

        next_retransmission = self._schedule_retransmit(message, timeout, 0)
        self._active_exchanges[key] = (messageerror_monitor, next_retransmission)

        self.log.debug("Exchange added, message ID: %d.", message.mid)

    def _remove_exchange(self, message):
        """Remove exchange from active exchanges and cancel the timeout to next
        retransmission."""
        key = (message.remote, message.mid)

        if key not in self._active_exchanges:
            # Before turning this up to a warning, consider https://github.com/chrysn/aiocoap/issues/288
            self.log.info("Received %s from %s, but could not match it to a running exchange.", message.mtype, message.remote)
            return

        messageerror_monitor, next_retransmission = self._active_exchanges.pop(key)
        next_retransmission.cancel()
        if message.mtype is RST:
            messageerror_monitor()
        self.log.debug("Exchange removed, message ID: %d.", message.mid)

        self._continue_backlog(message.remote)

    def _continue_backlog(self, remote):
        """After an exchange has been removed, start working off the backlog or
        clear it completely."""

        if remote not in self._backlogs:
            # if active exchanges were something we could do a
            # .register_finally() on, we could chain them like that; if we
            # implemented anything but NSTART=1, we'll need a more elaborate
            # system anyway
            raise AssertionError("backlogs/active_exchange relation violated (implementation error)")

        # first iteration is sure to happen, others happen only if the enqueued
        # messages were NONs
        while not any(r == remote for r, mid in self._active_exchanges.keys()):
            if self._backlogs[remote] != []:
                next_message, messageerror_monitor = self._backlogs[remote].pop(0)
                self._send_initially(next_message, messageerror_monitor)
            else:
                del self._backlogs[remote]
                break

    def _schedule_retransmit(self, message, timeout, retransmission_counter):
        """Create and return a call_later for first or subsequent
        retransmissions."""

        # while this could just as well be done in a lambda or with the
        # arguments passed to call_later, in this form makes the test cases
        # easier to debug (it's about finding where references to a Context
        # are kept around; contexts should be able to shut down in an orderly
        # way without littering references in the loop)

        def retr(self=self,
                message=message,
                timeout=timeout,
                retransmission_counter=retransmission_counter,
                doc="If you read this, have a look at _schedule_retransmit",
                id=object()):
            self._retransmit(message, timeout, retransmission_counter)
        return self.loop.call_later(timeout, retr)

    def _retransmit(self, message, timeout, retransmission_counter):
        """Retransmit CON message that has not been ACKed or RSTed."""
        key = (message.remote, message.mid)

        messageerror_monitor, next_retransmission = self._active_exchanges.pop(key)
        # this should be a no-op, but let's be sure
        next_retransmission.cancel()

        if retransmission_counter < message.transport_tuning.MAX_RETRANSMIT:
            self.log.info("Retransmission, Message ID: %d.", message.mid)
            self._send_via_transport(message)
            retransmission_counter += 1
            timeout *= 2

            next_retransmission = self._schedule_retransmit(message, timeout, retransmission_counter)
            self._active_exchanges[key] = (messageerror_monitor, next_retransmission)
        else:
            self.log.info("Exchange timed out trying to transmit %s", message)
            del self._backlogs[message.remote]
            self.token_manager.dispatch_error(error.ConRetransmitsExceeded("Retransmissions exceeded"), message.remote)

    #
    # coap dispatch, message-code sublayer: triggering custom actions based on incoming messages
    #

    def _process_ping(self, message):
        self.log.info('Received CoAP Ping from %s, replying with RST.', message.remote)
        rst = Message(mtype=RST, mid=message.mid, code=EMPTY, payload=b'')
        rst.remote = message.remote.as_response_address()
        # not going via send_message because that would strip the mid, and we
        # already know that it can go straight to the wire
        self._send_initially(rst)

    def _process_request(self, request):
        """Spawn a Responder for an incoming request, or feed a long-running
        responder if one exists."""

        if request.mtype == CON:
            def on_timeout(self, remote, token):
                mid, own_timeout = self._piggyback_opportunities.pop(
                        (remote, token))
                self._send_empty_ack(request.remote, mid,
                    "Response took too long to prepare")
            handle = self.loop.call_later(request.transport_tuning.EMPTY_ACK_DELAY,
                    on_timeout, self, request.remote, request.token)
            key = (request.remote, request.token)
            if key in self._piggyback_opportunities:
                self.log.warning("New request came in while old request not"
                      " ACKed yet. Possible mismatch between EMPTY_ACK_DELAY"
                      " and EXCHANGE_LIFETIME. Cancelling ACK to ward off any"
                      " further confusion.")
                mid, old_handle = self._piggyback_opportunities.pop(key)
                old_handle.cancel()
            self._piggyback_opportunities[key] = (request.mid, handle)

        self.token_manager.process_request(request)

    def _process_response(self, response):
        """Feed a response back to whatever might expect it.

        Returns True if the response was expected (and should be ACK'd
        depending on mtype), and False if it was not expected (and should be
        RST'd)."""

        self.log.debug("Received Response: %r", response)

        return self.token_manager.process_response(response)

    #
    # outgoing messages
    #

    async def fill_or_recognize_remote(self, message):
        if message.remote is not None:
            if await self.message_interface.recognize_remote(message.remote):
                return True
        remote = await self.message_interface.determine_remote(message)
        if remote is not None:
            message.remote = remote
            return True
        return False

    def send_message(self, message, messageerror_monitor):
        """Encode and send message. This takes care of retransmissions (if
        CON), message IDs and rate limiting, but does not hook any events to
        responses. (Use the :class:`Request` class or responding resources
        instead; those are the typical callers of this function.)

        If notification about the progress of the exchange is required, an
        ExchangeMonitor can be passed in, which will receive the appropriate
        callbacks."""

        if message.mid is not None:
            # if you can give any reason why the application should provide a
            # fixed mid, lower the log level on demand and provide the reason
            # in a comment.
            self.log.warning("Message ID set on to-be-sent message, this is"
                  " probably unintended; clearing it.")
            message.mid = None

        if message.code.is_response():
            no_response = (message.opt.no_response or 0) & (1 << message.code.class_ - 1) != 0

            piggyback_key = (message.remote, message.token)
            if piggyback_key in self._piggyback_opportunities:
                mid, handle = self._piggyback_opportunities.pop(piggyback_key)
                handle.cancel()

                if no_response:
                    new_message = Message(code=EMPTY, mid=mid, mtype=ACK)
                    new_message.remote = message.remote.as_response_address()
                    message = new_message
                    self.log.debug("Turning to-be-sent message into an empty ACK due to no_response option.")
                else:
                    message.mtype = ACK
                    message.mid = mid
            else:
                if no_response:
                    self.log.debug("Stopping message in message manager as it is no_response and no ACK is pending.")
                    return

            message.opt.no_response = None

        if message.mtype is None:
            if self._active_exchanges is None:
                # during shutdown, this is all we can do
                message.mtype = NON
            else:
                if message.remote.is_multicast:
                    message.mtype = NON
                else:
                    # FIXME: on responses, this should take the request into
                    # consideration (cf. RFC7252 Section 5.2.3, answer to NON
                    # SHOULD be NON)
                    message.mtype = CON
        else:
            if self._active_exchanges is None:
                self.log.warning("Forcing message to be sent as NON even though specified because transport is shutting down")
                message.mtype = NON

        if message.mtype == CON and message.remote.is_multicast:
            raise ValueError("Refusing to send CON message to multicast address")

        if message.mid is None:
            message.mid = self._next_message_id()

        if message.mtype == CON and message.remote in self._backlogs:
            assert any(remote == message.remote for (remote, _) in self._active_exchanges)
            self.log.debug("Message to %s put into backlog", message.remote)
            self._backlogs[message.remote].append((message, messageerror_monitor))
        else:
            self._send_initially(message, messageerror_monitor)

    def _send_initially(self, message, messageerror_monitor=None):
        """Put the message on the wire for the first time, starting retransmission timeouts"""

        self.log.debug("Sending message %r", message)

        if message.mtype is CON:
            assert messageerror_monitor is not None, "messageerror_monitor needs to be set for CONs"
            self._add_exchange(message, messageerror_monitor)

        self._store_response_for_duplicates(message)

        self._send_via_transport(message)

    def _send_via_transport(self, message):
        """Put the message on the wire"""

        self.message_interface.send(message)

    def _next_message_id(self):
        """Reserve and return a new message ID."""
        message_id = self.message_id
        self.message_id = 0xFFFF & (1 + self.message_id)
        return message_id

    def _send_empty_ack(self, remote, mid, reason):
        """Send separate empty ACK for any reason.

        Currently, this can happen only once per Responder, that is, when the
        last block1 has been transferred and the first block2 is not ready
        yet."""

        self.log.debug("Sending empty ACK: %s", reason)
        ack = Message(
                mtype=ACK,
                code=EMPTY,
                payload=b"",
                )
        ack.remote = remote.as_response_address()
        ack.mid = mid
        # not going via send_message because that would strip the mid, and we
        # already know that it can go straight to the wire
        self._send_initially(ack)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import asyncio
from collections import namedtuple
import functools
import sys

from . import error
from .numbers import INTERNAL_SERVER_ERROR
from .util.asyncio import py38args

class Pipe:
    """Low-level meeting point between a request and a any responses that come
    back on it.

    A single request message is placed in the Pipe at creation time.
    Any responses, as well as any exception happening in the course of
    processing, are passed back to the requester along the Pipe. A
    response can carry an indication of whether it is final; an exception
    always is.

    This object is used both on the client side (where the Context on behalf of
    the application creates a Pipe and passes it to the network
    transports that send the request and fill in any responses) and on the
    server side (where the Context creates one for an incoming request and
    eventually lets the server implementation populate it with responses).

    This currently follows a callback dispatch style. (It may be developed into
    something where only awaiting a response drives the proces, though).

    Currently, the requester sets up the object, connects callbacks, and then
    passes the Pipe on to whatever creates the response.

    The creator of responses is notified by the Pipe of a loss of
    interest in a response when there are no more callback handlers registered
    by registering an on_interest_end callback. As the response callbacks need
    to be already in place when the Pipe is passed on to the
    responder, the absence event callbacks is signalled by callign the callback
    immediately on registration.

    To accurately model "loss of interest", it is important to use the
    two-phase setup of first registering actual callbacks and then producing
    events and/or placing on_interest_end callbacks; this is not clearly
    expressed in type or state yet. (One possibility would be for the
    Pipe to carry a preparation boolean, and which prohibits event
    sending during preparation and is_interest=True callback creation
    afterwards).

    This was previously named PlumbingRequest.

    **Stability**

    Sites and resources implemented by provinding a
    :meth:`~aiocoap.interfaces.Resource.render_to_pipe` method can stably use
    the :meth:`add_response` method of a Pipe (or something that quacks like
    it).

    They should not rely on :meth:`add_exception` but rather just raise the
    exception, and neither register :meth:`on_event` handlers (being the sole
    producer of events) nor hook to :meth:`on_interest_end` (instead, they can
    use finally clauses or async context managers to handle any cleanup when
    the cancellation of the render task indicates the peer's loss of interest).
    """

    Event = namedtuple("Event", ("message", "exception", "is_last"))

    # called by the initiator of the request

    def __init__(self, request, log):
        self.request = request
        self.log = log

        self._event_callbacks = []
        """list[(callback, is_interest)], or None during event processing, or
        False when there were no more event callbacks and an the
        on_interest_end callbacks have already been called"""

    def __repr__(self):
        return '<%s at %#x around %r with %r callbacks (thereof %r interests)>' % (
                type(self).__name__,
                id(self),
                self.request,
                len(self._event_callbacks) if self._event_callbacks else self._event_callbacks,
                sum(1 for (e, is_interest) in self._event_callbacks if is_interest) if self._event_callbacks else self._event_callbacks,
                )

    def _any_interest(self):
        return any(is_interest for (cb, is_interest) in self._event_callbacks)

    def poke(self):
        """Ask the responder for a life sign. It is up to the responder to
        ignore this (eg. because the responder is the library/application and
        can't be just gone), to issue a generic transport-dependent 'ping' to
        see whether the connection is still alive, or to retransmit the request
        if it is an observation over an unreliable channel.

        In any case, no status is reported directly to the poke, but if
        whatever the responder does fails, it will send an appropriate error
        message as a response."""
        raise NotImplementedError()

    def on_event(self, callback, is_interest=True):
        """Call callback on any event. The callback must return True to be
        called again after an event. Callbacks must not produce new events or
        deregister unrelated event handlers.

        If is_interest=False, the callback will not be counted toward the
        active callbacks, and will receive a (None, None, is_last=True) event
        eventually.

        To unregister the handler, call the returned closure; this can trigger
        on_interest_end callbacks.
        """
        self._event_callbacks.append((callback, is_interest))
        return functools.partial(self._unregister_on_event, callback)

    def _unregister_on_event(self, callback):
        if self._event_callbacks is False:
            # They wouldn't be called any more so they're already dropped.a
            # It's OK that the caller cleans up after itself: Sure it could
            # register an on_interest_end, but that's really not warranted if
            # all it wants to know is whether it'll have to execute cleanup
            # when it's shutting down or not.
            return

        self._event_callbacks = [(cb, i) for (cb, i) in self._event_callbacks if callback is not cb]
        if not self._any_interest():
            self._end()

    def on_interest_end(self, callback):
        """Register a callback that will be called exactly once -- either right
        now if there is not even a current indicated interest, or at a last
        event, or when no more interests are present"""

        if self._event_callbacks is False:
            # Happens, for example, when a proxy receives multiple requests on a single token
            self.log.warning("on_interest_end callback %r added after %r has already ended", callback, self)
            callback()
            return

        if self._any_interest():
            self._event_callbacks.append((
                lambda e: ((callback(), False) if e.is_last else (None, True))[1],
                False
                ))
        else:
            callback()

    def _end(self):
        cbs = self._event_callbacks
        self._event_callbacks = False
        tombstone = self.Event(None, None, True)
        [cb(tombstone) for (cb, _) in cbs]

    # called by the responding side

    def _add_event(self, event):
        if self._event_callbacks is False:
            # Happens, for example, when a proxy receives multiple requests on a single token
            self.log.warning("Response %r added after %r has already ended", event, self)
            return

        for (cb, is_interest) in self._event_callbacks[:]:
            keep_calling = cb(event)
            if not keep_calling:
                if self._event_callbacks is False:
                    # All interest was just lost during the callback
                    return

                self._event_callbacks.remove((cb, is_interest))

        if not self._any_interest():
            self._end()

    def add_response(self, response, is_last=False):
        self._add_event(self.Event(response, None, is_last))

    def add_exception(self, exception):
        self._add_event(self.Event(None, exception, True))

def run_driving_pipe(pipe, coroutine, name=None):
    """Create a task from a coroutine where the end of the coroutine produces a
    terminal event on the pipe, and lack of interest in the pipe cancels the
    task.

    The coroutine will typically produce output into the pipe; that
    connection is set up by the caller like as in
    ``run_driving_pipe(pipe, render_to(pipe))``.

    The create task is not returned, as the only sensible operation on it would
    be cancellation and that's already set up from the pipe.
    """

    async def wrapped():
        try:
            await coroutine
        except Exception as e:
            pipe.add_exception(e)
        # Not doing anything special about cancellation: it indicates the
        # peer's loss of interest, so there's no use in sending anythign out to
        # someone not listening any more
        #
        # (We'd *like* to do something on the Python 3.7 versions whose
        # cancelled threads show errors, but there we can't stop it in here
        # because catching the CancelledError doesn't remove the taint from the
        # task).

    task = asyncio.create_task(
            wrapped(),
            **py38args(name=name),
            )
    if sys.version_info < (3, 8):
        # These Python versions used to complain about cancelled tasks, where
        # really a cancelled task is perfectly natural (especially here where
        # it's just not needed any more because nobody is listening to what it
        # produces). As catching CancellationError doesn't help silencing them,
        # this workaround ensures the cancellations don't raise.
        def silence_cancellation(task):
            try:
                task.result()
            except asyncio.CancelledError:
                pass
        task.add_done_callback(silence_cancellation)
    pipe.on_interest_end(task.cancel)

def error_to_message(old_pr, log):
    """Given a pipe set up by the requester, create a new pipe to pass on to a
    responder.

    Any exceptions produced by the responder will be turned into terminal
    responses on the original pipe, and loss of interest is forwarded."""

    from .message import Message

    next_pr = Pipe(old_pr.request, log)

    def on_event(event):
        if event.message is not None:
            old_pr.add_response(event.message, event.is_last)
            return not event.is_last

        e = event.exception

        if isinstance(e, error.RenderableError):
            # the repr() here is quite imporant for garbage collection
            log.info("Render request raised a renderable error (%s), responding accordingly.", repr(e))
            try:
                msg = e.to_message()
                if msg is None:
                    # This deserves a separate check because the ABC checks
                    # that should ensure that the default to_message method is
                    # never used in concrete classes fails due to the metaclass
                    # conflict between ABC and Exceptions
                    raise ValueError("Exception to_message failed to produce a message on %r" % e)
            except Exception as e2:
                log.error("Rendering the renderable exception failed: %r", e2, exc_info=e2)
                msg = Message(code=INTERNAL_SERVER_ERROR)
            old_pr.add_response(msg, is_last=True)
        else:
            log.error("An exception occurred while rendering a resource: %r", e, exc_info=e)
            old_pr.add_response(Message(code=INTERNAL_SERVER_ERROR), is_last=True)

        return False

    remove_interest = next_pr.on_event(on_event)
    old_pr.on_interest_end(remove_interest)
    return next_pr

class IterablePipe:
    """A stand-in for a Pipe that the requesting party can use
    instead. It should behave just like a Pipe to the responding
    party, but the caller does not register on_event handlers and instead
    iterates asynchronously over the events.

    Note that the PR can be aitered over only once, and does not support any
    additional hook settings once asynchronous iteration is started; this is
    consistent with the usage pattern of pipes.
    """

    def __init__(self, request):
        self.request = request

        self.__on_interest_end = []

        # FIXME: This is unbounded -- pipes should gain support for
        # backpressure.
        self.__queue = asyncio.Queue()

    def on_interest_end(self, callback):
        try:
            self.__on_interest_end.append(callback)
        except AttributeError:
            raise RuntimeError("Attempted to declare interest in the end of a IterablePipe on which iteration already started") from None

    def __aiter__(self):
        i = self.Iterator(self.__queue, self.__on_interest_end)
        del self.__on_interest_end
        return i

    def _add_event(self, e):
        self.__queue.put_nowait(e)

    def add_response(self, response, is_last=False):
        self._add_event(Pipe.Event(response, None, is_last))

    def add_exception(self, exception):
        self._add_event(Pipe.Event(None, exception, True))

    class Iterator:
        def __init__(self, queue, on_interest_end):
            self.__queue = queue
            self.__on_interest_end = on_interest_end

        async def __anext__(self):
            return await self.__queue.get()

        def __del__(self):
            # This is pretty reliable as the iterator is only created and
            # referenced in the desugaring of the `async for`.
            for c in self.__on_interest_end:
                c()



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""
Workaround for https://github.com/pyca/cryptography/issues/5557

These functions could be methods to
`cryptography.hazmat.primitives.asymmetric.ed25519.{Ed25519PrivateKey,
Ed25519PublicKey}`, respectively, and are currently implemented manually or
using ge25519.

These conversions are not too critical in that they do not run on data an
attacker can send arbitrarily (in the most dynamic situation, the keys are
distributed through a KDC aka. group manager).
"""

from cryptography.hazmat.primitives.asymmetric import ed25519, x25519
from cryptography.hazmat.primitives import serialization

def sk_to_curve25519(ed: ed25519.Ed25519PrivateKey) -> x25519.X25519PrivateKey:
    raw = ed.private_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PrivateFormat.Raw,
            encryption_algorithm=serialization.NoEncryption(),
            )

    # as proposed in https://github.com/pyca/cryptography/issues/5557#issuecomment-739339132

    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.backends.openssl.backend import backend

    hasher = hashes.Hash(hashes.SHA512())
    hasher.update(raw)
    h = bytearray(hasher.finalize())
    # curve25519 clamping
    h[0] &= 248
    h[31] &= 127
    h[31] |= 64

    return backend.x25519_load_private_bytes(h[0:32])

def pk_to_curve25519(ed: ed25519.Ed25519PublicKey) -> x25519.X25519PublicKey:
    raw = ed.public_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw,
            )

    # This is libsodium's crypto_sign_ed25519_pk_to_curve25519 translated into
    # the Pyton module ge25519.

    from ge25519 import ge25519, ge25519_p3
    from fe25519 import fe25519

    if ge25519.has_small_order(raw) != 0:
        raise RuntimeError("Doesn' thave small order")

    # frombytes in libsodium appears to be the same as
    # frombytes_negate_vartime; as ge25519 only implements the from_bytes
    # version, we have to do the root check manually.
    A = ge25519_p3.from_bytes(raw)
    if A.root_check:
        raise RuntimeError("Root check failed")

    if not A.is_on_main_subgroup():
        raise RuntimeError("It's on the main subgroup")

    one_minus_y = fe25519.one() - A.Y
    x = A.Y + fe25519.one()
    x = x * one_minus_y.invert()

    return x25519.X25519PublicKey.from_public_bytes(bytes(x.to_bytes()))



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""aiocoap-client is a simple command-line tool for interacting with CoAP servers"""

import sys
import asyncio
import argparse
import logging
import subprocess
from pathlib import Path

import shlex
# even though not used directly, this has side effects on the input() function
# used in interactive mode
try:
    import readline # noqa: F401
except ImportError:
    pass # that's normal on some platforms, and ok since it's just a usability enhancement

import aiocoap
import aiocoap.defaults
import aiocoap.meta
import aiocoap.proxy.client
from aiocoap.util import contenttype
from aiocoap.util.cli import ActionNoYes
from aiocoap.numbers import ContentFormat
from ..util.asyncio import py38args

def build_parser():
    p = argparse.ArgumentParser(description=__doc__)
    p.add_argument('--non', help="Send request as non-confirmable (NON) message", action='store_true')
    p.add_argument('-m', '--method', help="Name or number of request method to use (default: %(default)s)", default="GET")
    p.add_argument('--observe', help="Register an observation on the resource", action='store_true')
    p.add_argument('--observe-exec', help="Run the specified program whenever the observed resource changes, feeding the response data to its stdin", metavar='CMD')
    p.add_argument('--accept', help="Content format to request", metavar="MIME")
    p.add_argument('--proxy', help="Relay the CoAP request to a proxy for execution", metavar="URI")
    p.add_argument('--payload', help="Send X as request payload (eg. with a PUT). If X starts with an '@', its remainder is treated as a file name and read from; '@-' reads from the console. Non-file data may be recoded, see --content-format.", metavar="X")
    p.add_argument('--payload-initial-szx', help="Size exponent to limit the initial block's size (0 ≙ 16 Byte, 6 ≙ 1024 Byte)", metavar="SZX", type=int)
    p.add_argument('--content-format', help="Content format of the --payload data. If a known format is given and --payload has a non-file argument, the payload is converted from CBOR Diagnostic Notation.", metavar="MIME")
    p.add_argument('--no-set-hostname', help="Suppress transmission of Uri-Host even if the host name is not an IP literal", dest="set_hostname", action='store_false', default=True)
    p.add_argument('-v', '--verbose', help="Increase the debug output", action="count")
    p.add_argument('-q', '--quiet', help="Decrease the debug output", action="count")
    p.add_argument('--interactive', help="Enter interactive mode", action="store_true") # careful: picked before parsing
    p.add_argument('--credentials', help="Load credentials to use from a given file", type=Path)
    p.add_argument('--version', action="version", version='%(prog)s ' + aiocoap.meta.version)

    p.add_argument('--color',
            help="Color output (default on TTYs if all required modules are installed)",
            default=None,
            action=ActionNoYes,
            )
    p.add_argument('--pretty-print',
            help="Pretty-print known content formats (default on TTYs if all required modules are installed)",
            default=None,
            action=ActionNoYes,
            )
    p.add_argument('url', help="CoAP address to fetch")

    return p

def configure_logging(verbosity):
    logging.basicConfig()

    if verbosity <= -2:
        logging.getLogger('coap').setLevel(logging.CRITICAL + 1)
    elif verbosity == -1:
        logging.getLogger('coap').setLevel(logging.ERROR)
    elif verbosity == 0:
        logging.getLogger('coap').setLevel(logging.WARNING)
    elif verbosity == 1:
        logging.getLogger('coap').setLevel(logging.INFO)
    elif verbosity >= 2:
        logging.getLogger('coap').setLevel(logging.DEBUG)

def colored(text, options, *args, **kwargs):
    """Apply termcolor.colored with the given args if options.color is set"""
    if not options.color:
        return text

    import termcolor
    return termcolor.colored(text, *args, **kwargs)

def incoming_observation(options, response):
    if options.observe_exec:
        p = subprocess.Popen(options.observe_exec, shell=True, stdin=subprocess.PIPE)
        # FIXME this blocks
        p.communicate(response.payload)
    else:
        sys.stdout.write(colored('---', options, 'grey', attrs=['bold']) + '\n')
        if response.code.is_successful():
            present(response, options, file=sys.stderr)
        else:
            sys.stdout.flush()
            print(colored(response.code, options, 'red'), file=sys.stderr)
            if response.payload:
                present(response, options, file=sys.stderr)

def apply_credentials(context, credentials, errfn):
    if credentials.suffix == '.json':
        import json
        context.client_credentials.load_from_dict(json.load(credentials.open('rb')))
    else:
        raise errfn("Unknown suffix: %s (expected: .json)" % (credentials.suffix))

def present(message, options, file=sys.stdout):
    """Write a message payload to the output, pretty printing and/or coloring
    it as configured in the options."""
    if not options.quiet and (message.opt.location_path or message.opt.location_query):
        # FIXME: Percent encoding is completely missing; this would be done
        # most easily with a CRI library
        location_ref = "/" + "/".join(message.opt.location_path)
        if message.opt.location_query:
            location_ref += "?" + "&".join(message.opt.location_query)
        print(colored(f"Location options indicate new resource: {location_ref}", options, 'green'), file=sys.stderr)

    if not message.payload:
        return

    payload = None

    cf = message.opt.content_format or message.request.opt.content_format
    if cf is not None and cf.is_known():
        mime = cf.media_type
    else:
        mime = 'application/octet-stream'
    if options.pretty_print:
        from aiocoap.util.prettyprint import pretty_print
        prettyprinted = pretty_print(message)
        if prettyprinted is not None:
            (infos, mime, payload) = prettyprinted
            if not options.quiet:
                for i in infos:
                    print(colored(i, options, 'grey', attrs=['bold']),
                            file=sys.stderr)

    color = options.color
    if color:
        from aiocoap.util.prettyprint import lexer_for_mime
        import pygments
        try:
            lexer = lexer_for_mime(mime)
        except pygments.util.ClassNotFound:
            color = False

    if color and payload is None:
        # Coloring requires a unicode-string style payload, either from the
        # mime type or from the pretty printer.
        try:
            payload = message.payload.decode('utf8')
        except UnicodeDecodeError:
            color = False

    if color:
        from pygments.formatters import TerminalFormatter
        from pygments import highlight
        highlit = highlight(
            payload,
            lexer,
            TerminalFormatter(),
            )
        # The TerminalFormatter already adds an end-of-line character, not
        # trying to add one for any missing trailing newlines.
        print(highlit, file=file, end="")
        file.flush()
    else:
        if payload is None:
            file.buffer.write(message.payload)
            if file.isatty() and message.payload[-1:] != b'\n':
                file.write("\n")
        else:
            file.write(payload)
            if file.isatty() and payload[-1] != '\n':
                file.write("\n")

async def single_request(args, context):
    parser = build_parser()
    options = parser.parse_args(args)

    pretty_print_modules = aiocoap.defaults.prettyprint_missing_modules()
    if pretty_print_modules and \
            (options.color is True or options.pretty_print is True):
        parser.error("Color and pretty printing require the following"
                " additional module(s) to be installed: %s" %
                ", ".join(pretty_print_modules))
    if options.color is None:
        options.color = sys.stdout.isatty() and not pretty_print_modules
    if options.pretty_print is None:
        options.pretty_print = sys.stdout.isatty() and not pretty_print_modules

    configure_logging((options.verbose or 0) - (options.quiet or 0))

    try:
        code = getattr(
                aiocoap.numbers.codes.Code,
                options.method.upper().replace('IPATCH', 'iPATCH')
                )
    except AttributeError:
        try:
            code = aiocoap.numbers.codes.Code(int(options.method))
        except ValueError:
            raise parser.error("Unknown method")

    if options.credentials is not None:
        apply_credentials(context, options.credentials, parser.error)

    request = aiocoap.Message(code=code, mtype=aiocoap.NON if options.non else aiocoap.CON)
    try:
        request.set_request_uri(options.url, set_uri_host=options.set_hostname)
    except ValueError as e:
        raise parser.error(e)

    if not request.opt.uri_host and not request.unresolved_remote:
        raise parser.error("Request URLs need to be absolute.")

    if options.accept:
        try:
            request.opt.accept = ContentFormat(int(options.accept))
        except ValueError:
            try:
                request.opt.accept = ContentFormat.by_media_type(options.accept)
            except KeyError:
                raise parser.error("Unknown accept type")

    if options.observe:
        request.opt.observe = 0
        observation_is_over = asyncio.get_event_loop().create_future()

    if options.content_format:
        try:
            request.opt.content_format = ContentFormat(int(options.content_format))
        except ValueError:
            try:
                request.opt.content_format = ContentFormat.by_media_type(options.content_format)
            except KeyError:
                raise parser.error("Unknown content format")

    if options.payload:
        if options.payload.startswith('@'):
            filename = options.payload[1:]
            if filename == "-":
                f = sys.stdin.buffer
            else:
                f = open(filename, 'rb')
            try:
                request.payload = f.read()
            except OSError as e:
                raise parser.error("File could not be opened: %s"%e)
        else:
            request_classification = contenttype.categorize(
                    request.opt.content_format.media_type
                    if request.opt.content_format is not None and
                        request.opt.content_format.is_known()
                    else ""
                    )
            if request_classification in ('cbor', 'cbor-seq'):
                try:
                    import cbor_diag
                except ImportError as e:
                    raise parser.error(f"CBOR recoding not available ({e})")

                try:
                    encoded = cbor_diag.diag2cbor(options.payload)
                except ValueError as e:
                    raise parser.error(f"Parsing CBOR diagnostic notation failed. Make sure quotation marks are escaped from the shell. Error: {e}")

                if request_classification == 'cbor-seq':
                    try:
                        import cbor2
                    except ImportError as e:
                        raise parser.error(f"CBOR sequence recoding not available ({e})")
                    decoded = cbor2.loads(encoded)
                    if not isinstance(decoded, list):
                        raise parser.error("CBOR sequence recoding requires an array as the top-level element.")
                    request.payload = b"".join(cbor2.dumps(d) for d in decoded)
                else:
                    request.payload = encoded
            else:
                request.payload = options.payload.encode('utf8')

    if options.payload_initial_szx is not None:
        request.opt.block1 = aiocoap.optiontypes.BlockOption.BlockwiseTuple(
                0,
                False,
                options.payload_initial_szx,
            )

    if options.proxy is None:
        interface = context
    else:
        interface = aiocoap.proxy.client.ProxyForwarder(options.proxy, context)

    try:
        requested_uri = request.get_request_uri()

        requester = interface.request(request)

        if options.observe:
            requester.observation.register_errback(observation_is_over.set_result)
            requester.observation.register_callback(lambda data, options=options: incoming_observation(options, data))

        try:
            response_data = await requester.response
        except aiocoap.error.ResolutionError as e:
            print("Name resolution error:", e, file=sys.stderr)
            sys.exit(1)
        except aiocoap.error.NetworkError as e:
            print("Network error:", e, file=sys.stderr)
            sys.exit(1)
        # Fallback while not all backends raise NetworkErrors
        except OSError as e:
            text = str(e)
            if not text:
                text = repr(e)
            if not text:
                # eg ConnectionResetError flying out of a misconfigured SSL server
                text = type(e)
            print("Error:", text, file=sys.stderr)
            sys.exit(1)

        response_uri = response_data.get_request_uri()
        if requested_uri != response_uri:
            print("Response arrived from different address; base URI is",
                    response_uri, file=sys.stderr)
        if response_data.code.is_successful():
            present(response_data, options)
        else:
            print(colored(response_data.code, options, 'red'), file=sys.stderr)
            present(response_data, options, file=sys.stderr)
            sys.exit(1)

        if options.observe:
            exit_reason = await observation_is_over
            print("Observation is over: %r"%(exit_reason,), file=sys.stderr)
    finally:
        if not requester.response.done():
            requester.response.cancel()
        if options.observe and not requester.observation.cancelled:
            requester.observation.cancel()

async def single_request_with_context(args):
    """Wrapper around single_request until sync_main gets made fully async, and
    async context managers are used to manage contexts."""
    context = await aiocoap.Context.create_client_context()
    try:
        await single_request(args, context)
    finally:
        await context.shutdown()

interactive_expecting_keyboard_interrupt = None

async def interactive():
    global interactive_expecting_keyboard_interrupt
    interactive_expecting_keyboard_interrupt = asyncio.get_event_loop().create_future()

    context = await aiocoap.Context.create_client_context()

    while True:
        try:
            # when http://bugs.python.org/issue22412 is resolved, use that instead
            line = await asyncio.get_event_loop().run_in_executor(None, lambda: input("aiocoap> "))
        except EOFError:
            line = "exit"
        line = shlex.split(line)
        if not line:
            continue
        if line in (["help"], ["?"]):
            line = ["--help"]
        if line in (["quit"], ["q"], ["exit"]):
            break

        current_task = asyncio.create_task(
                single_request(line, context=context),
                **py38args(name="Interactive prompt command %r" % line)
                )
        interactive_expecting_keyboard_interrupt = asyncio.get_event_loop().create_future()

        done, pending = await asyncio.wait([current_task, interactive_expecting_keyboard_interrupt], return_when=asyncio.FIRST_COMPLETED)

        if current_task not in done:
            current_task.cancel()
        else:
            try:
                await current_task
            except SystemExit as e:
                if e.code != 0:
                    print("Exit code: %d"%e.code, file=sys.stderr)
                continue
            except Exception as e:
                print("Unhandled exception raised: %s"%(e,))

    await context.shutdown()

def sync_main(args=None):
    # interactive mode is a little messy, that's why this is not using aiocoap.util.cli yet
    if args is None:
        args = sys.argv[1:]

    if '--interactive' not in args:
        try:
            asyncio.run(single_request_with_context(args))
        except KeyboardInterrupt:
            sys.exit(3)
    else:
        if len(args) != 1:
            print("No other arguments must be specified when entering interactive mode", file=sys.stderr)
            sys.exit(1)

        loop = asyncio.get_event_loop()
        task = loop.create_task(
                interactive(),
                **py38args(name="Interactive prompt")
                )

        while not task.done():
            try:
                loop.run_until_complete(task)
            except KeyboardInterrupt:
                if not interactive_expecting_keyboard_interrupt.done():
                    interactive_expecting_keyboard_interrupt.set_result(None)
            except SystemExit:
                continue # asyncio/tasks.py(242) raises those after setting them as results, but we particularly want them back in the loop

if __name__ == "__main__":
    sync_main()



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Tools that I'd like to have in urllib.parse"""

import string

#: "unreserved" characters from RFC3986
unreserved = string.ascii_letters + string.digits + '-._~'

#: "sub-delims" characters from RFC3986
sub_delims = "!$&'()*+,;="

def quote_factory(safe_characters):
    """Return a quote function that escapes all characters not in the
    safe_characters iterable."""
    safe_set = set(ord(x) for x in safe_characters)
    if any(c >= 128 for c in safe_set):
        raise ValueError("quote_factory does not support non-ASCII safe characters")
    def quote(input_string):
        encoded = input_string.encode('utf8')
        return "".join(chr(x) if x in safe_set else "%%%02X" % x for x in encoded)
    return quote



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""List of known values for the CoAP "Code" field.

The values in this module correspond to the IANA registry "`CoRE Parameters`_",
subregistries "CoAP Method Codes" and "CoAP Response Codes".

The codes come with methods that can be used to get their rough meaning, see
the :class:`Code` class for details.

.. _`CoRE Parameters`: https://www.iana.org/assignments/core-parameters/core-parameters.xhtml
"""

import warnings
import random

from ..util import ExtensibleIntEnum

class Code(ExtensibleIntEnum):
    """Value for the CoAP "Code" field.

    As the number range for the code values is separated, the rough meaning of
    a code can be determined using the :meth:`is_request`, :meth:`is_response` and
    :meth:`is_successful` methods."""

EMPTY = 0
    GET = 1
    POST = 2
    PUT = 3
    DELETE = 4
    FETCH = 5
    PATCH = 6
    iPATCH = 7

    # RESPONSE CODES
    CREATED = 131#65
    DELETED = 65#66
    VALID = 133#67
    CHANGED = 69#68
    CONTENT = 140#69
    CONTINUE = 130#95

    BAD_REQUEST = 141#128
    UNAUTHORIZED = 162#129
    BAD_OPTION = 161#130
    FORBIDDEN = 163#131
    NOT_FOUND = 68#132
    METHOD_NOT_ALLOWED = 164#133
    NOT_ACCEPTABLE = 132#134
    PRECONDITION_FAILED = 95#140
    REQUEST_ENTITY_TOO_LARGE = 160#141
    UNSUPPORTED_CONTENT_FORMAT = 134#143
    INTERNAL_SERVER_ERROR = 66#160
    NOT_IMPLEMENTED = 143#161
    BAD_GATEWAY = 165#162
    SERVICE_UNAVAILABLE = 67#163
    GATEWAY_TIMEOUT = 128#164
    PROXYING_NOT_SUPPORTED = 129#165
    #------------------------------------
    
    REQUEST_ENTITY_INCOMPLETE = 136
    CONFLICT = (4 << 5) + 9
    @property
    def UNSUPPORTED_MEDIA_TYPE(self):
        warnings.warn("UNSUPPORTED_MEDIA_TYPE is a deprecated alias for UNSUPPORTED_CONTENT_FORMAT")
        return self.UNSUPPORTED_CONTENT_FORMAT
    UNPROCESSABLE_ENTITY = (4 << 5) + 22
    TOO_MANY_REQUESTS = (4 << 5) + 29
    HOP_LIMIT_REACHED = (5 << 5) + 8

    CSM = 225
    PING = 226
    PONG = 227
    RELEASE = 228
    ABORT = 229

    def is_request(self):
        """True if the code is in the request code range"""
        return True if (self >= 1 and self < 32) else False


    def is_response(self):
        """True if the code is in the response code range"""
        return True if (self >= 64 and self < 192) else False

    def is_signalling(self):
        return True if self >= 224 else False


    def is_successful(self):
        """True if the code is in the successful subrange of the response code range"""
        return True if (self >= 64 and self < 96) else False

    def can_have_payload(self):
        """True if a message with that code can carry a payload. This is not
        checked for strictly, but used as an indicator."""
        return self.is_response() or self in (self.POST, self.PUT, self.FETCH, self.PATCH, self.iPATCH)

    @property
    def class_(self):
        """The class of a code (distinguishing whether it's successful, a
        request or a response error or more).

        >>> Code.CONTENT
        <Successful Response Code 69 "2.05 Content">
        >>> Code.CONTENT.class_
        2
        >>> Code.BAD_GATEWAY
        <Response Code 162 "5.02 Bad Gateway">
        >>> Code.BAD_GATEWAY.class_
        5
        """
        return self >> 5

    @property
    def dotted(self):
        """The numeric value three-decimal-digits (c.dd) form"""
        return "%d.%02d" % divmod(self, 32)

    @property
    def name_printable(self):
        """The name of the code in human-readable form"""
        return self.name.replace('_', ' ').title()

    def __str__(self):
        """
        >>> print(Code.GET)
        GET
        >>> print(Code.CONTENT)
        2.05 Content
        >>> print(Code.BAD_GATEWAY)
        5.02 Bad Gateway
        >>> print(Code(32))
        32
        """
        if self.is_request() or self is self.EMPTY:
            return self.name
        elif self.is_response() or self.is_signalling():
            return "%s %s" % (self.dotted, self.name_printable)
        else:
            return "%d" % self

    def _classification(self):
        return ("Successful " if self.is_successful() else "") + ("Request " if self.is_request() else "Response " if self.is_response() else "")

    def __repr__(self):
        """
        >>> Code.GET
        <Request Code 1 "GET">
        >>> Code.CONTENT
        <Successful Response Code 69 "2.05 Content">
        >>> Code.BAD_GATEWAY
        <Response Code 162 "5.02 Bad Gateway">
        >>> Code(32)
        <Code 32 "32">
        """
        return '<%sCode %d "%s">' % (self._classification(), self, self)

    name = property(lambda self: self._name if hasattr(self, "_name") else "(unknown)", lambda self, value: setattr(self, "_name", value), doc="The constant name of the code (equals name_printable readable in all-caps and with underscores)")

    def _repr_html_(self):
        import html
        if hasattr(self, "_name"):
            return f'<abbr title="{self._classification()}Code {self.dotted}">{html.escape(self.name)}</abbr>'
        else:
            return f'<abbr title="Unknown {self._classification()}Code">{self.dotted}</abbr>'

for k in vars(Code):
    if isinstance(getattr(Code, k), Code):
        locals()[k] = getattr(Code, k)

__all__ = ['Code'] + [k for (k, v) in locals().items() if isinstance(v, Code)]



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Known values for CoAP option numbers

The values defined in `OptionNumber` correspond to the IANA registry "CoRE
Parameters", subregistries "CoAP Method Codes" and "CoAP Response Codes".

The option numbers come with methods that can be used to evaluate their
properties, see the `OptionNumber` class for details.
"""

from ..util import ExtensibleIntEnum
from .. import optiontypes

class OptionNumber(ExtensibleIntEnum):
    """A CoAP option number.

    As the option number contains information on whether the option is
    critical, and whether it is safe-to-forward, those properties can be
    queried using the `is_*` group of methods.

    Note that whether an option may be repeated or not does not only depend on
    the option, but also on the context, and is thus handled in the `Options`
    object instead."""
    IF_MATCH = 1
    URI_HOST = 3
    ETAG = 4
    IF_NONE_MATCH = 5
    OBSERVE = 6
    URI_PORT = 7
    LOCATION_PATH = 8
    # The OBJECT_SECURITY name is to be deprecated when Python 3.8 support is
    # dropped, and we can do that easily using a class property.
    OSCORE = OBJECT_SECURITY = 9
    URI_PATH = 11
    CONTENT_FORMAT = 12
    MAX_AGE = 14
    URI_QUERY = 15
    HOP_LIMIT = 16
    ACCEPT = 17
    Q_BLOCK1 = 19
    LOCATION_QUERY = 20
    EDHOC = 21
    BLOCK2 = 23
    BLOCK1 = 27
    SIZE2 = 28
    Q_BLOCK2 = 31
    PROXY_URI = 35
    PROXY_SCHEME = 39
    SIZE1 = 60
    ECHO = 252
    NO_RESPONSE = 258
    REQUEST_TAG = 292

    # experimental for draft-amsuess-core-cachable-oscore
    #
    # Using the number suggested there (rather than a high one) as this is
    # going to be used in overhead comparisons.
    REQUEST_HASH = 548

    def __add__(self, delta):
        """Addition makes sense on these due to the delta encoding in CoAP
        serialization"""
        return type(self)(int(self) + delta)

    def is_critical(self):
        return self & 0x01 == 0x01

    def is_elective(self):
        return not self.is_critical()

    def is_unsafe(self):
        return self & 0x02 == 0x02

    def is_safetoforward(self):
        return not self.is_unsafe()

    def is_nocachekey(self):
        if self.is_unsafe():
            raise ValueError("NoCacheKey is only meaningful for safe options")
        return self & 0x1e == 0x1c

    def is_cachekey(self):
        return not self.is_nocachekey()

    def _get_format(self):
        if hasattr(self, "_format"):
            return self._format
        else:
            return optiontypes.OpaqueOption

    def _set_format(self, value):
        self._format = value

    format = property(_get_format, _set_format)

    def create_option(self, decode=None, value=None):
        """Return an Option element of the appropriate class from this option
        number.

        An initial value may be set using the decode or value options, and will
        be fed to the resulting object's decode method or value property,
        respectively."""
        option = self.format(self)
        if decode is not None:
            option.decode(decode)
        if value is not None:
            option.value = value
        return option

    def _repr_html_(self):
        import html
        properties = f"{'critical' if self.is_critical() else 'elective'}, {'safe-to-forward' if self.is_safetoforward() else 'proxy unsafe'}"
        if self.is_safetoforward():
            properties += ", part of the cache key" if self.is_cachekey() else ", not part of the cache key"
        if hasattr(self, "name"):
            return f'<abbr title="option {int(self)}: {properties}">{html.escape(self.name)}</abbr>'
        else:
            return f'<abbr title="{properties}">Option {int(self)}</abbr>'

# OpaqueOption is set on formats where it is known to be used even though it is
# the default. This allows developers to rely on those interfaces to be stable
# (or at least to be notified visibly in the release notes).

# RFC 7252

OptionNumber.IF_MATCH.format = optiontypes.OpaqueOption
OptionNumber.URI_HOST.format = optiontypes.StringOption
OptionNumber.ETAG.format = optiontypes.OpaqueOption
OptionNumber.URI_PORT.format = optiontypes.UintOption
OptionNumber.LOCATION_PATH.format = optiontypes.StringOption
OptionNumber.URI_PATH.format = optiontypes.StringOption
OptionNumber.CONTENT_FORMAT.format = optiontypes.ContentFormatOption
OptionNumber.MAX_AGE.format = optiontypes.UintOption
OptionNumber.URI_QUERY.format = optiontypes.StringOption
OptionNumber.ACCEPT.format = optiontypes.ContentFormatOption
OptionNumber.LOCATION_QUERY.format = optiontypes.StringOption
OptionNumber.PROXY_URI.format = optiontypes.StringOption
OptionNumber.PROXY_SCHEME.format = optiontypes.StringOption
OptionNumber.SIZE1.format = optiontypes.UintOption

# RFC 7959

OptionNumber.BLOCK2.format = optiontypes.BlockOption
OptionNumber.BLOCK1.format = optiontypes.BlockOption
OptionNumber.SIZE2.format = optiontypes.UintOption

# RFC 7641

OptionNumber.OBSERVE.format = optiontypes.UintOption

# RFC 7967

OptionNumber.NO_RESPONSE.format = optiontypes.UintOption

# RFC 8613

OptionNumber.OSCORE.format = optiontypes.OpaqueOption

# RFC 9175

OptionNumber.ECHO.format = optiontypes.OpaqueOption
OptionNumber.REQUEST_TAG.format = optiontypes.OpaqueOption

# RFC 8768

OptionNumber.HOP_LIMIT.format = optiontypes.UintOption

# experimental for draft-amsuess-core-cachable-oscore

OptionNumber.REQUEST_HASH.format = optiontypes.OpaqueOption



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""a plain CoAP proxy that can work both as forward and as reverse proxy"""

import sys
import argparse

import aiocoap
from aiocoap.proxy.server import ForwardProxyWithPooledObservations, ProxyWithPooledObservations, NameBasedVirtualHost, SubdomainVirtualHost, SubresourceVirtualHost, UnconditionalRedirector
from aiocoap.util.cli import AsyncCLIDaemon
from aiocoap.cli.common import add_server_arguments, server_context_from_arguments

def build_parser():
    p = argparse.ArgumentParser(description=__doc__)

    mode = p.add_argument_group("mode", "Required argument for setting the operation mode")
    mode.add_argument('--forward', help="Run as forward proxy", action='store_true')
    mode.add_argument('--reverse', help="Run as reverse proxy", action='store_true')

    details = p.add_argument_group("details", "Options that govern how requests go in and out")
    add_server_arguments(details)
    details.add_argument("--register", help="Register with a Resource directory", metavar='RD-URI', nargs='?', default=False)
    details.add_argument("--register-as", help="Endpoint name (with possibly a domain after a dot) to register as", metavar='EP[.D]', default=None)
    details.add_argument("--register-proxy", help="Ask the RD to serve as a reverse proxy. Note that this is only practical for --unconditional or --pathbased reverse proxies.", action='store_true')

    r = p.add_argument_group('Rules', description="Sequence of forwarding rules "
            "that, if matched by a request, specify a forwarding destination. Destinations can be prefixed to change their behavior: With an '@' sign, they are treated as forward proxies. With a '!' sign, the destination is set as Uri-Host.")
    class TypedAppend(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            if getattr(namespace, self.dest) is None:
                setattr(namespace, self.dest, [])
            getattr(namespace, self.dest).append((option_string, values))
    r.add_argument('--namebased', help="If Uri-Host matches NAME, route to DEST", metavar="NAME:DEST", action=TypedAppend, dest='r')
    r.add_argument('--subdomainbased', help="If Uri-Host is anything.NAME, route to DEST", metavar="NAME:DEST", action=TypedAppend, dest='r')
    r.add_argument('--pathbased', help="If a requested path starts with PATH, split that part off and route to DEST", metavar="PATH:DEST", action=TypedAppend, dest='r')
    r.add_argument('--unconditional', help="Route all requests not previously matched to DEST", metavar="DEST", action=TypedAppend, dest='r')

    return p

def destsplit(dest):
    use_as_proxy = False
    rewrite_uri_host = False
    if dest.startswith('!'):
        dest = dest[1:]
        rewrite_uri_host = True
    if dest.startswith('@'):
        dest = dest[1:]
        use_as_proxy = True
    return dest, rewrite_uri_host, use_as_proxy

class Main(AsyncCLIDaemon):
    async def start(self, args=None):
        parser = build_parser()
        options = parser.parse_args(args if args is not None else sys.argv[1:])
        self.options = options

        if not options.forward and not options.reverse:
            raise parser.error("At least one of --forward and --reverse must be given.")

        self.outgoing_context = await aiocoap.Context.create_client_context()
        if options.forward:
            proxy = ForwardProxyWithPooledObservations(self.outgoing_context)
        else:
            proxy = ProxyWithPooledObservations(self.outgoing_context)
        for kind, data in options.r or ():
            if kind in ('--namebased', '--subdomainbased'):
                try:
                    name, dest = data.split(':', 1)
                except Exception:
                    raise parser.error("%s needs NAME:DEST as arguments" % kind)
                dest, rewrite_uri_host, use_as_proxy = destsplit(dest)
                if rewrite_uri_host and kind == '--subdomainbased':
                    parser.error("The flag '!' makes no sense for subdomain based redirection as the subdomain data would be lost")
                r = (NameBasedVirtualHost if kind == '--namebased' else SubdomainVirtualHost)(name, dest, rewrite_uri_host, use_as_proxy)
            elif kind == '--pathbased':
                try:
                    path, dest = data.split(':', 1)
                except Exception:
                    raise parser.error("--pathbased needs PATH:DEST as arguments")
                r = SubresourceVirtualHost(path.split('/'), dest)
            elif kind == '--unconditional':
                dest, rewrite_uri_host, use_as_proxy = destsplit(data)
                if rewrite_uri_host:
                    parser.error("The flag '!' makes no sense for unconditional redirection as the host name data would be lost")
                r = UnconditionalRedirector(dest, use_as_proxy)
            else:
                raise AssertionError('Unknown redirectory kind')
            proxy.add_redirector(r)

        self.proxy_context = await server_context_from_arguments(proxy, options)

        if options.register is not False:
            from aiocoap.resourcedirectory.client.register import Registerer

            params = {}
            if options.register_as:
                ep, _, d = options.register_as.partition('.')
                params['ep'] = ep
                if d:
                    params['d'] = d
            if options.register_proxy:
                # FIXME: Check this in discovery
                params['proxy'] = 'on'
            # FIXME: Construct this from settings (path-based), and forward results
            proxy.get_resources_as_linkheader = lambda: ""
            self.registerer = Registerer(self.proxy_context, rd=options.register, lt=60,
                    registration_parameters=params)

    async def shutdown(self):
        if self.options.register is not False:
            await self.registerer.shutdown()
        await self.outgoing_context.shutdown()
        await self.proxy_context.shutdown()

sync_main = Main.sync_main

if __name__ == "__main__":
    # if you want to run this using `python3 -m`, see http://bugs.python.org/issue22480
    sync_main()



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This module implements a MessageInterface that serves coaps:// using a
wrapped tinydtls library.

Bear in mind that the aiocoap CoAPS support is highly experimental and
incomplete.

Unlike other transports this is *not* enabled automatically in general, as it
is limited to servers bound to a single address for implementation reasons.
(Basically, because it is built on the simplesocketserver rather than the udp6
server -- that can change in future, though). Until either the implementation
is changed or binding arguments are (allowing different transports to bind to
per-transport addresses or ports), a DTLS server will only be enabled if the
AIOCOAP_DTLSSERVER_ENABLED environment variable is set, or tinydtls_server is
listed explicitly in AIOCOAP_SERVER_TRANSPORT.
"""

# Comparing this to the tinydtls transport, things are a bit easier as we don't
# expect to send the first DTLS payload (thus don't need the queue), and don't
# need that clean a cleanup (at least if we assume that the clients all shut
# down on their own anyway).
#
# Then again, keeping connections live for as long as someone holds their
# address (eg. by some "pool with N strong references, and the rest are weak"
# and just go away on overflow unless someone keeps the address alive) would be
# more convenient here.

import asyncio
from collections import OrderedDict

import time

from ..numbers.constants import COAPS_PORT
from .generic_udp import GenericMessageInterface
from .. import error, interfaces
from . import simplesocketserver
from .simplesocketserver import _DatagramServerSocketSimple
from ..util.asyncio import py38args

from .tinydtls import LEVEL_NOALERT, LEVEL_FATAL, DTLS_EVENT_CONNECT, DTLS_EVENT_CONNECTED, CODE_CLOSE_NOTIFY, CloseNotifyReceived, DTLS_TICKS_PER_SECOND, DTLS_CLOCK_OFFSET, FatalDTLSError

# tinyDTLS passes address information around in its session data, but the way
# it's used here that will be ignored; this is the data that is sent to / read
# from the tinyDTLS functions
_SENTINEL_ADDRESS = "::1"
_SENTINEL_PORT = 1234

# While we don't have retransmissions set up, this helps work issues of dropped
# packets from sending in rapid succession
_SEND_SLEEP_WORKAROUND = 0

class _AddressDTLS(interfaces.EndpointAddress):
    # no slots here, thus no equality other than identity, which is good

    def __init__(self, protocol, underlying_address):
        from DTLSSocket import dtls

        self._protocol = protocol
        self._underlying_address = simplesocketserver._Address(protocol, underlying_address)

        self._dtls_socket = None

        self._psk_store = SecurityStore(protocol._server_credentials)

        self._dtls_socket = dtls.DTLS(
                # FIXME: Use accessors like tinydtls (but are they needed? maybe shutdown sequence is just already better here...)
                read=self._read,
                write=self._write,
                event=self._event,
                pskId=b"The socket needs something there but we'll never use it",
                pskStore=self._psk_store,
                )
        self._dtls_session = dtls.Session(_SENTINEL_ADDRESS, _SENTINEL_PORT)

        self._retransmission_task = asyncio.create_task(
                self._run_retransmissions(),
                **py38args(name="DTLS server handshake retransmissions")
                )

        self.log = protocol.log

    is_multicast = False
    is_multicast_locally = False
    hostinfo = property(lambda self: self._underlying_address.hostinfo)
    uri_base = property(lambda self: 'coaps://' + self.hostinfo)
    hostinfo_local = property(lambda self: self._underlying_address.hostinfo_local)
    uri_base_local = property(lambda self: 'coaps://' + self.hostinfo_local)

    scheme = 'coaps'

    authenticated_claims = property(lambda self: [self._psk_store._claims])

    @property
    def blockwise_key(self):
        return (self._underlying_address.blockwise_key, self._psk_store._claims)

    # implementing GenericUdp addresses

    def send(self, message):
        self._dtls_socket.write(self._dtls_session, message)

    # dtls callbacks

    def _read(self, sender, data):
        # ignoring sender: it's only _SENTINEL_*
        self._protocol._message_interface._received_plaintext(self, data)

        return len(data)

    def _write(self, recipient, data):
        if _SEND_SLEEP_WORKAROUND and \
                len(data) > 13 and data[0] == 22 and data[13] == 14:
            time.sleep(_SEND_SLEEP_WORKAROUND)
        self._underlying_address.send(data)
        return len(data)

    def _event(self, level, code):
        if (level, code) == (LEVEL_NOALERT, DTLS_EVENT_CONNECT):
            return
        elif (level, code) == (LEVEL_NOALERT, DTLS_EVENT_CONNECTED):
            # No need to react to "connected": We're not the ones sending the first message
            return
        elif (level, code) == (LEVEL_FATAL, CODE_CLOSE_NOTIFY):
            self._inject_error(CloseNotifyReceived())
        elif level == LEVEL_FATAL:
            self._inject_error(FatalDTLSError(code))
        else:
            self.log.warning("Unhandled alert level %d code %d", level, code)

    # own helpers copied and adjusted from tinydtls

    def _inject_error(self, e):
        # this includes "was shut down" with a CloseNotifyReceived e
        """Put an error to all pending operations on this remote, just as if it
        were raised inside the main loop."""
        self._protocol._message_interface._received_exception(self, e)

        self._retransmission_task.cancel()

        self._protocol._connections.pop(self._underlying_address.address)

    # This is a bit more defensive than the one in tinydtls as it starts out in
    # waiting, and RFC6347 indicates on a brief glance that the state machine
    # could go from waiting to some other state later on, so we (re)trigger it
    # whenever something comes in
    async def _run_retransmissions(self):
        when = self._dtls_socket.checkRetransmit() / DTLS_TICKS_PER_SECOND
        if when == 0:
            return
        # FIXME: Find out whether the DTLS server is ever supposed to send
        # retransmissions in the first place (this part was missing an import
        # and it never showed).
        now = time.time() - DTLS_CLOCK_OFFSET
        await asyncio.sleep(when - now)
        self._retransmission_task = asyncio.create_task(
                self._run_retransmissions(),
                **py38args(name="DTLS server handshake retransmissions")
                )

class _DatagramServerSocketSimpleDTLS(_DatagramServerSocketSimple):
    _Address = _AddressDTLS
    max_sockets = 64

    def __init__(self, *args, **kwargs):
        self._connections = OrderedDict() # analogous to simple6's _sockets
        return super().__init__(*args, **kwargs)

    async def connect(self, sockaddr):
        # Even if we opened a connection, it wouldn't have the same security
        # properties as the incoming one that it's probably supposed to replace
        # would have had
        raise RuntimeError("Sending initial messages via a DTLSServer is not supported")

    # Overriding to use GoingThroughMessageDecryption adapter
    @classmethod
    async def create(cls, bind, log, loop, message_interface):
        wrapped_interface = GoingThroughMessageDecryption(message_interface)
        self = await super().create(bind, log, loop, wrapped_interface)
        # self._security_store left uninitialized to ease subclassing from SimpleSocketServer; should be set before using this any further
        return self

    # Overriding as now we do need to manage the pol
    def datagram_received(self, data, sockaddr):
        if sockaddr in self._connections:
            address = self._connections[sockaddr]
            self._connections.move_to_end(sockaddr)
        else:
            address = self._Address(self, sockaddr)
            self._connections[sockaddr] = address
        self._message_interface._received_datagram(address, data)

    def _maybe_purge_sockets(self):
        while len(self._connections) >= self.max_sockets: # more of an if
            oldaddr, oldest = next(iter(self._connections.items()))
            # FIXME custom error?
            oldest._inject_error(error.LibraryShutdown("Connection is being closed for lack of activity"))

class GoingThroughMessageDecryption:
    """Warapper around GenericMessageInterface that puts incoming data through
    the DTLS context stored with the address"""
    def __init__(self, plaintext_interface: "GenericMessageInterface"):
        self._plaintext_interface = plaintext_interface

    def _received_datagram(self, address, data):
        # Put it into the DTLS processor; that'll forward any actually contained decrypted datagrams on to _received_plaintext
        address._retransmission_task.cancel()
        address._dtls_socket.handleMessage(address._dtls_session, data)
        address._retransmission_task = asyncio.create_task(
                address._run_retransmissions(),
                **py38args(name="DTLS server handshake retransmissions")
                )

    def _received_exception(self, address, exception):
        self._plaintext_interface._received_exception(address, exception)

    def _received_plaintext(self, address, data):
        self._plaintext_interface._received_datagram(address, data)

class SecurityStore:
    """Wrapper around a CredentialsMap that makes it accessible to the
    dict-like object DTLSSocket expects.

    Not only does this convert interfaces, it also adds a back channel: As
    DTLSSocket wouldn't otherwise report who authenticated, this is tracking
    access and storing the claims associated with the used key for later use.

    Therefore, SecurityStore objects are created per connection and not per
    security store.
    """

    def __init__(self, server_credentials):
        self._server_credentials = server_credentials

        self._claims = None

    def keys(self):
        return self

    def __contains__(self, key):
        try:
            self._server_credentials.find_dtls_psk(key)
            return True
        except KeyError:
            return False

    def __getitem__(self, key):
        (psk, claims) = self._server_credentials.find_dtls_psk(key)
        if self._claims not in (None, claims):
            # I didn't know it could do that -- how would we know which is the
            # one it eventually picked?
            raise RuntimeError("DTLS stack tried accessing different keys")
        self._claims = claims
        return psk

class MessageInterfaceTinyDTLSServer(simplesocketserver.MessageInterfaceSimpleServer):
    _default_port = COAPS_PORT
    _serversocket = _DatagramServerSocketSimpleDTLS

    @classmethod
    async def create_server(cls, bind, ctx: interfaces.MessageManager, log, loop, server_credentials):
        self = await super().create_server(bind, ctx, log, loop)

        self._pool._server_credentials = server_credentials

        return self

    async def shutdown(self):
        remaining_connections = list(self._pool._connections.values())
        for c in remaining_connections:
            c._inject_error(error.LibraryShutdown("Shutting down"))
        await super().shutdown()



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Container module for transports

Transports are expected to be the modular backends of aiocoap, and implement
the specifics of eg. TCP, WebSockets or SMS, possibly divided by backend
implementations as well.

Transports are not part of the API, so the class descriptions in the modules
are purely informational.

Multiple transports can be used in parallel in a single :class:`.Context`, and
are loaded in a particular sequence. Some transports will grab all addresses of
a given protocol, so they might not be practical to combine. Which transports
are started in a given Context follows the
:func:`.defaults.get_default_clienttransports` function.

The available transports are:

.. the files in this directory.
"""



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

# WORK IN PROGRESS: TransportEndpoint has been renamed to MessageInterface
# here, but actually we'll be providing a RequestInterface -- that's one of the
# reasons why RequestInterface, TokenInterface and MessageInterface were split
# in the first place.

"""This module implements a RequestProvider for OSCORE. As such, it takes
routing ownership of requests that it has a security context available for, and
sends off the protected messages via another transport.

This transport is a bit different from the others because it doesn't have its
dedicated URI scheme, but purely relies on preconfigured contexts.

So far, this transport only deals with outgoing requests, and does not help in
building an OSCORE server. (Some code that could be used here in future resides
in `contrib/oscore-plugtest/plugtest-server` as the `ProtectedSite` class.

In outgoing request, this transport automatically handles Echo options that
appear to come from RFC8613 Appendix B.1.2 style servers. They indicate that
the server could not process the request initially, but could do so if the
client retransmits it with an appropriate Echo value.

Unlike other transports that could (at least in theory) be present multiple
times in :attr:`aiocoap.protocol.Context.request_interfaces` (eg. because there
are several bound sockets), this is only useful once in there, as it has no own
state, picks the OSCORE security context from the CoAP
:attr:`aiocoap.protocol.Context.client_credentials` when populating the remote
field, and handles any populated request based ono its remote.security_context
property alone.
"""

from collections import namedtuple
from functools import wraps

from .. import interfaces, credentials, oscore
from ..numbers import UNAUTHORIZED, MAX_REGULAR_BLOCK_SIZE_EXP
from ..util.asyncio import py38args

class OSCOREAddress(
        namedtuple("_OSCOREAddress", ["security_context", "underlying_address"]),
        interfaces.EndpointAddress
        ):
    """Remote address type for :class:`TransportOSCORE`."""

    def __repr__(self):
        return "<%s in context %r to %r>" % (type(self).__name__, self.security_context, self.underlying_address)

    def _requires_ua(f):
        @wraps(f)
        def wrapper(self):
            if self.underlying_address is None:
                raise ValueError("No underlying address populated that could be used to derive a hostinfo")
            return f(self)
        return wrapper

    @property
    @_requires_ua
    def hostinfo(self):
        return self.underlying_address.hostinfo

    @property
    @_requires_ua
    def hostinfo_local(self):
        return self.underlying_address.hostinfo_local

    @property
    @_requires_ua
    def uri_base(self):
        return self.underlying_address.uri_base

    @property
    @_requires_ua
    def uri_base_local(self):
        return self.underlying_address.uri_base_local

    @property
    @_requires_ua
    def scheme(self):
        return self.underlying_address.scheme

    @property
    def authenticated_claims(self):
        return self.security_context.authenticated_claims

    is_multicast = False

    maximum_payload_size = 1024
    maximum_block_size_exp = MAX_REGULAR_BLOCK_SIZE_EXP

    @property
    def blockwise_key(self):
        if hasattr(self.security_context, 'groupcontext'):
            # it's an aspect, and all aspects work compatibly as long as data
            # comes from the same recipient ID -- taking the group recipient
            # key for that one which is stable across switches between pairwise
            # and group mode
            detail = self.security_context.groupcontext.recipient_keys[self.security_context.recipient_id]
        else:
            detail = self.security_context.recipient_key
        return (self.underlying_address.blockwise_key, detail)

class TransportOSCORE(interfaces.RequestProvider):
    def __init__(self, context, forward_context):
        self._context = context
        self._wire = forward_context

        if self._context.loop is not self._wire.loop:
            # TransportOSCORE is not designed to bridge loops -- would probably
            # be possible, but incur confusion that is most likely well avoidable
            raise ValueError("Wire and context need to share an asyncio loop")

        self.loop = self._context.loop
        self.log = self._context.log

        # Keep current requests. This is not needed for shutdown purposes (see
        # .shutdown), but because Python 3.6.4 (but not 3.6.5, and not at least
        # some 3.5) would otherwise cancel OSCORE tasks mid-observation. This
        # manifested itself as <https://github.com/chrysn/aiocoap/issues/111>.
        self._tasks = set()

    #
    # implement RequestInterface
    #

    async def fill_or_recognize_remote(self, message):
        if isinstance(message.remote, OSCOREAddress):
            return True
        if message.opt.object_security is not None:
            # double oscore is not specified; using this fact to make `._wire
            # is ._context` an option
            return False

        try:
            secctx = self._context.client_credentials.credentials_from_request(message)
        except credentials.CredentialsMissingError:
            return False

        # FIXME: it'd be better to have a "get me credentials *of this type* if they exist"
        if isinstance(secctx, oscore.CanProtect):
            message.remote = OSCOREAddress(secctx, message.remote)
            self.log.debug("Selecting OSCORE transport based on context %r for new request %r", secctx, message)
            return True
        else:
            return False

    def request(self, request):
        t = self.loop.create_task(
                self._request(request),
                **py38args(name="OSCORE request %r" % request)
                )
        self._tasks.add(t)
        t.add_done_callback(lambda _, _tasks=self._tasks, _t=t: _tasks.remove(_t))

    async def _request(self, request):
        msg = request.request

        secctx = msg.remote.security_context

        def protect(echo):
            if echo is None:
                msg_to_protect = msg
            else:
                if msg.opt.echo:
                    self.log.warning("Overwriting the requested Echo value with the one to answer a 4.01 Unauthorized")
                msg_to_protect = msg.copy(echo=echo)
            protected, original_request_seqno = secctx.protect(msg_to_protect)
            protected.remote = msg.remote.underlying_address

            wire_request = self._wire.request(protected)

            return (wire_request, original_request_seqno)

        wire_request, original_request_seqno = protect(None)

        # tempting as it would be, we can't access the request as a
        # Pipe here, because it is a BlockwiseRequest to handle
        # outer blockwise.
        # (Might be a good idea to model those after Pipe too,
        # though).

        def _check(more, unprotected_response):
            if more and not unprotected_response.code.is_successful():
                self.log.warning("OSCORE protected message contained observe, but unprotected code is unsuccessful. Ignoring the observation.")
                return False
            return more

        try:
            protected_response = await wire_request.response

            # Offer secctx to switch over for reception based on the header
            # data (similar to how the server address switches over when
            # receiving a response to a request sent over multicast)
            unprotected = oscore.verify_start(protected_response)
            secctx = secctx.context_from_response(unprotected)

            unprotected_response, _ = secctx.unprotect(protected_response, original_request_seqno)

            if unprotected_response.code == UNAUTHORIZED and unprotected_response.opt.echo is not None:
                # Assist the server in B.1.2 Echo receive window recovery
                self.log.info("Answering the server's 4.01 Unauthorized / Echo as part of OSCORE B.1.2 recovery")

                wire_request, original_request_seqno = protect(unprotected_response.opt.echo)

                protected_response = await wire_request.response
                unprotected_response, _ = secctx.unprotect(protected_response, original_request_seqno)

            unprotected_response.remote = OSCOREAddress(secctx, protected_response.remote)
            self.log.debug("Successfully unprotected %r into %r", protected_response, unprotected_response)
            # FIXME: if i could tap into the underlying Pipe, that'd
            # be a lot easier -- and also get rid of the awkward _check
            # code moved into its own function just to avoid duplication.
            more = protected_response.opt.observe is not None
            more = _check(more, unprotected_response)
            request.add_response(unprotected_response, is_last=not more)

            if not more:
                return

            async for protected_response in wire_request.observation:
                unprotected_response, _ = secctx.unprotect(protected_response, original_request_seqno)

                more = protected_response.opt.observe is not None
                more = _check(more, unprotected_response)

                unprotected_response.remote = OSCOREAddress(secctx, protected_response.remote)
                self.log.debug("Successfully unprotected %r into %r", protected_response, unprotected_response)
                # FIXME: discover is_last from the underlying response
                request.add_response(unprotected_response, is_last=not more)

                if not more:
                    return
            request.add_exception(NotImplementedError("End of observation"
                " should have been indicated in is_last, see above lines"))
        except Exception as e:
            request.add_exception(e)
        finally:
            # FIXME: no way yet to cancel observations exists yet, let alone
            # one that can be used in a finally clause (ie. won't raise
            # something else if the observation terminated server-side)
            pass
            #if wire_request.observation is not None:
            #    wire_request.observation.cancel()

    async def shutdown(self):
        # Nothing to do here yet; the individual requests will be shut down by
        # their underlying transports
        pass



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This module describes how security credentials are expressed in aiocoap,
how security protocols (TLS, DTLS, OSCOAP) can store and access their key
material, and for which URIs they are used.

For consistency, mappings between accessible resources and their credentials
are always centered around URIs. This is slightly atypical, because a client
will typically use a particular set of credentials for all operations on one
server, while a server first loads all available credentials and then filters
out whether the client may actually access a resource per-path, but it works
with full URIs (or patterns thereof) just as well. That approach allows using
more similar structures both on the server and the client, and works smoothly
for virtual hosting, firewalling and clients accessing resources with varying
credentials.

Still, client and server credentials are kept apart, lest a server open up (and
potentially reveal) to a PSK set it is only configured to use as a client.
While client credentials already have their place in
:attr:`aiocoap.protocol.Context.client_credentials`, server credentials are not
in use at a standardized location yet because there is only code in the OSCORE
plug tests that can use it so far.

Library developer notes
~~~~~~~~~~~~~~~~~~~~~~~

This whole module currently relies on a mixture of introspection and manual
parsing of the JSON-ish tree. A preferred expression of the same would rely on
the credentials.cddl description and build an object tree from that, but the
author is unaware of any existing CDDL Python implementation. That might also
ease porting to platforms that don't support inspect like micropython does.
"""

import re
import inspect

from typing import Optional


'''
server: {
            'coaps://mysite/*': { 'dtls-psk' (or other granularity): { 'psk': 'abcd' }},
            'coap://mysite/*': { 'oscore': { 'contextfile': 'my-contextfile/' } },
            'coap://myothersite/firmware': ':myotherkey',
            'coap://myothersite/reset': ':myotherkey',
            'coap://othersite*': { 'unprotected': true },
            ':myotherkey': { 'oscore': { 'contextfile': 'my-contextfile/' } }
        }

server can of course just say it doesn't want to have the Site handle it and
just say '*': { 'unprotected': true }, add some ':foo': {'dtls-psk': ...}
entries (so communication can be established in the first place) and let
individual resources decide whether they return 4.01 or something else.

client can be the same with different implied role, or have something like

client: {
            'coap://myothersite/*': ':myotherkey',
            ...
        }

in future also

server: {
        'coaps://mysite/*': { 'dtls-cert': {'key': '...pem', 'cert': '...crt'} }
        }

client: {
        '*': { 'dtls-cert': { 'ca': '/etc/ssl/...' } }
}

or more complex ones:

server: {
        'coaps://myothersite/wellprotected': { 'all': [ ':mydtls', ':myotherkey' ]}
        'coaps://myothersite/*': { 'any': [ ':mydtls', ':myotherkey' ]}
}
'''

class CredentialsLoadError(ValueError):
    """Raised by functions that create a CredentialsMap or its parts from
    simple data structures"""

class CredentialsMissingError(RuntimeError):
    """Raised when no suiting credentials can be found for a message, or
    credentials are found but inapplicable to a transport's security
    mechanisms."""

class CredentialReference:
    def __init__(self, target, map):
        if not target.startswith(':'):
            raise CredentialsLoadError("Credential references must start with a colon (':')")
        self.target = target
        self.map = map

    # FIXME either generalize this with getattr, or introduce a function to
    # resolve any indirect credentials to a particular instance.

    def as_dtls_psk(self):
        return self.map[self.target].as_dtls_psk()

class _Listish(list):
    @classmethod
    def from_item(cls, v):
        if not isinstance(v, list):
            raise CredentialsLoadError("%s goes with a list" % cls.__name__)
        return cls(v)

class AnyOf(_Listish):
    pass

class AllOf(_Listish):
    pass

def _call_from_structureddata(constructor, name, init_data):
    if not isinstance(init_data, dict):
        raise CredentialsLoadError("%s goes with an object" % name)

    init_data = {k.replace('-', '_'): v for (k, v) in init_data.items()}

    sig = inspect.signature(constructor)

    checked_items = {}

    for k, v in init_data.items():
        try:
            annotation = sig.parameters[k].annotation
        except KeyError:
            # let this raise later in binding
            checked_items[k] = object()

        if isinstance(v, dict) and 'ascii' in v:
            if len(v) != 1:
                raise CredentialsLoadError("ASCII objects can only have one elemnt.")
            try:
                v = v['ascii'].encode('ascii')
            except UnicodeEncodeError:
                raise CredentialsLoadError("Elements of the ASCII object can not be represented in ASCII, please use binary or hex representation.")


        if isinstance(v, dict) and 'hex' in v:
            if len(v) != 1:
                raise CredentialsLoadError("Hex objects can only have one elemnt.")
            try:
                v = bytes.fromhex(v['hex'].replace('-', '').replace(' ', '').replace(':', ''))
            except ValueError as e:
                raise CredentialsLoadError("Hex object can not be read: %s" % (e.args[0]))

        # Not using isinstance because I foundno way to extract the type
        # information from an Optional/Union again; this whole thing works
        # only for strings and ints anyway, so why not.
        if annotation not in (type(v), Optional[type(v)]):
            # explicitly not excluding inspect._empty here: constructors
            # need to be fully annotated
            raise CredentialsLoadError("Type mismatch in attribute %s of %s: expected %s, got %r" % (k, name, annotation, v))

        checked_items[k] = v

    try:
        bound = sig.bind(**checked_items)
    except TypeError as e:
        raise CredentialsLoadError("%s: %s" % (name, e.args[0]))

    return constructor(*bound.args, **bound.kwargs)

class _Objectish:
    @classmethod
    def from_item(cls, init_data):
        return _call_from_structureddata(cls, cls.__name__, init_data)

class DTLS(_Objectish):
    def __init__(self, psk: bytes, client_identity: bytes):
        self.psk = psk
        self.client_identity = client_identity

    def as_dtls_psk(self):
        return (self.client_identity, self.psk)

class TLSCert(_Objectish):
    """Indicates that a client can use the given certificate file to authenticate the server.

    Can only be used with 'coaps+tcp://HOSTINFO/*' and 'coaps+tcp://*' forms.
    """
    def __init__(self, certfile: str):
        self.certfile = certfile

    def as_ssl_params(self):
        """Generate parameters suitable for passing via ** to
        ssl.create_default_context when purpose is alreay set"""
        return {"cafile": self.certfile}

def construct_oscore(contextfile: str):
    from .oscore import FilesystemSecurityContext

    return FilesystemSecurityContext(contextfile)

construct_oscore.from_item = lambda value: _call_from_structureddata(construct_oscore, 'OSCORE', value)

_re_cache = {}

class CredentialsMap(dict):
    """
    FIXME: outdated, rewrite when usable

    A CredentialsMap, for any URI template and operation, which
    security contexts are sufficient to to perform the operation on a matching
    URI.

    The same context can be used both by the server and the client, where the
    client uses the information on allowed client credentials to decide which
    credentials to present, and the information on allowed server credentials
    to decide whether the server can be trusted.

    Conversely, the server typically loads all available server credentials at
    startup, and then uses the client credentials list to decide whether to
    serve the request."""

    def load_from_dict(self, d):
        """Populate the map from a dictionary, which would typically have been
        loaded from a JSON/YAML file and needs to match the CDDL in
        credentials.cddl.

        Running this multiple times will overwriter individual entries in the
        map."""
        for k, v in d.items():
            if v is None:
                if k in self:
                    del self[k]
            else:
                self[k] = self._item_from_dict(v)
                # FIXME only works that way for OSCORE clients
                self[k].authenticated_claims = [k]

    def _item_from_dict(self, v):
        if isinstance(v, str):
            return CredentialReference(v, self)
        elif isinstance(v, dict):
            try:
                (key, value), = v.items()
            except ValueError:
                # this follows how Rust Enums are encoded in serde JSON
                raise CredentialsLoadError(
                        "Items in a credentials map must have exactly one key"
                        " (found %s)" % ("," .join(v.keys()) or "empty")
                    )

            try:
                constructor = self._class_map[key].from_item
            except KeyError:
                raise CredentialsLoadError("Unknown credential type: %s" % key)

            return constructor(value)

    _class_map = {
            'dtls': DTLS,
            'oscore': construct_oscore,
            'tlscert': TLSCert,
            'any-of': AnyOf,
            'all-of': AllOf,
            }

    @staticmethod
    def _wildcard_match(searchterm, pattern):
        if pattern not in _re_cache:
            _re_cache[pattern] = re.compile(re.escape(pattern).replace('\\*', '.*'))
        return _re_cache[pattern].fullmatch(searchterm) is not None

    # used by a client

    def credentials_from_request(self, msg):
        """Return the most specific match to a request message. Matching is
        currently based on wildcards, but not yet very well thought out."""

        uri = msg.get_request_uri()

        for i in range(1000):
            for (k, v) in sorted(self.items(), key=lambda x: len(x[0]), reverse=True):
                if self._wildcard_match(uri, k):
                    if isinstance(v, str):
                        uri = v
                        continue
                    return v
            else:
                raise CredentialsMissingError("No suitable credentials for %s" % uri)
        else:
            raise CredentialsLoadError("Search for suitable credentials for %s exceeds recursion limit")

    def ssl_client_context(self, scheme, hostinfo):
        """Return an SSL client context as configured for the given request
        scheme and hostinfo (no full message is to be processed here, as
        connections are used across requests to the same origin).

        If no credentials are configured, this returns None (for which the user
        may need to fill in ssl.create_default_context() if None is not already
        a good indicator for the eventual consumer to use the default)."""

        ssl_params = {}
        tlscert = self.get('%s://%s/*' % (scheme, hostinfo), None)
        if tlscert is None:
            tlscert = self.get('%s://*' % scheme, None)
        if tlscert is not None:
            ssl_params = tlscert.as_ssl_params()
        if ssl_params:
            import ssl
            return ssl.create_default_context(**ssl_params)

    # used by a server

    def find_oscore(self, unprotected):
        # FIXME: this is not constant-time as it should be, but too much in
        # flux to warrant optimization

        # FIXME: duplicate contexts for being tried out are not supported yet.

        for item in self.values():
            if not hasattr(item, "get_oscore_context_for"):
                continue

            ctx = item.get_oscore_context_for(unprotected)
            if ctx is not None:
                return ctx

        raise KeyError()

    def find_dtls_psk(self, identity):
        # FIXME similar to find_oscore
        for (entry, item) in self.items():
            if not hasattr(item, "as_dtls_psk"):
                continue

            psk_id, psk = item.as_dtls_psk()
            if psk_id != identity:
                continue

            # FIXME is returning the entry name a sane value to later put in to
            # authenticated_claims? OSCORE does something different.
            return (psk, entry)

        raise KeyError()



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This module implements a MessageInterface that handles coaps:// using a
wrapped tinydtls library.

This currently only implements the client side. To have a test server, run::

    $ git clone https://github.com/obgm/libcoap.git --recursive
    $ cd libcoap
    $ ./autogen.sh
    $ ./configure --with-tinydtls --disable-shared --disable-documentation
    $ make
    $ ./examples/coap-server -k secretPSK

(Using TinyDTLS in libcoap is important; with the default OpenSSL build, I've
seen DTLS1.0 responses to DTLS1.3 requests, which are hard to debug.)

The test server with its built-in credentials can then be accessed using::

    $ echo '{"coaps://localhost/*": {"dtls": {"psk": {"ascii": "secretPSK"}, "client-identity": {"ascii": "client_Identity"}}}}' > testserver.json
    $ ./aiocoap-client coaps://localhost --credentials testserver.json

While it is planned to allow more programmatical construction of the
credentials store, the currently recommended way of storing DTLS credentials is
to load a structured data object into the client_credentials store of the context:

>>> c = await aiocoap.Context.create_client_context()          # doctest: +SKIP
>>> c.client_credentials.load_from_dict(
...     {'coaps://localhost/*': {'dtls': {
...         'psk': b'secretPSK',
...         'client-identity': b'client_Identity',
...         }}})                                               # doctest: +SKIP

where, compared to the JSON example above, byte strings can be used directly
rather than expressing them as 'ascii'/'hex' (`{'hex': '30383135'}` style works
as well) to work around JSON's limitation of not having raw binary strings.

Bear in mind that the aiocoap CoAPS support is highly experimental; for
example, while requests to this server do complete, error messages are still
shown during client shutdown.
"""

import asyncio
import weakref
import functools
import time
import warnings

from ..util import hostportjoin, hostportsplit
from ..util.asyncio import py38args
from ..message import Message
from .. import interfaces, error
from ..numbers import COAPS_PORT
from ..credentials import CredentialsMissingError

# tinyDTLS passes address information around in its session data, but the way
# it's used here that will be ignored; this is the data that is sent to / read
# from the tinyDTLS functions
_SENTINEL_ADDRESS = "::1"
_SENTINEL_PORT = 1234

DTLS_EVENT_CONNECT = 0x01DC
DTLS_EVENT_CONNECTED = 0x01DE
DTLS_EVENT_RENEGOTIATE = 0x01DF

LEVEL_NOALERT = 0 # seems only to be issued by tinydtls-internal events

# from RFC 5246
LEVEL_WARNING = 1
LEVEL_FATAL = 2
CODE_CLOSE_NOTIFY = 0

# tinydtls can not be debugged in the Python way; if you need to get more
# information out of it, use the following line:
#dtls.setLogLevel(dtls.DTLS_LOG_DEBUG)

# FIXME this should be exposed by the dtls wrapper
DTLS_TICKS_PER_SECOND = 1000
DTLS_CLOCK_OFFSET = time.time()

# Currently kept a bit private by not inheriting from NetworkError -- thus
# they'll be wrapped in a NetworkError when they fly out of a request.
class CloseNotifyReceived(Exception):
    """The DTLS connection a request was sent on raised was closed by the
    server while the request was being processed"""

class FatalDTLSError(Exception):
    """The DTLS connection a request was sent on raised a fatal error while the
    request was being processed"""

class DTLSClientConnection(interfaces.EndpointAddress):
    # FIXME not only does this not do error handling, it seems not to even
    # survive its 2**16th message exchange.

    is_multicast = False
    is_multicast_locally = False
    hostinfo = None # stored at initualization time
    uri_base = property(lambda self: 'coaps://' + self.hostinfo)
    # Not necessarily very usable given we don't implement responding to server
    # connection, but valid anyway
    uri_base_local = property(lambda self: 'coaps://' + self.hostinfo_local)
    scheme = 'coaps'

    @property
    def hostinfo_local(self):
        # See TCP's.hostinfo_local
        host, port, *_ = self._transport.get_extra_info('socket').getsockname()
        if port == COAPS_PORT:
            port = None
        return hostportjoin(host, port)

    @property
    def blockwise_key(self):
        return (self._host, self._port, self._pskId, self._psk)

    def __init__(self, host, port, pskId, psk, coaptransport):
        self._ready = False
        self._queue = [] # stores sent packages while connection is being built

        self._host = host
        self._port = port
        self._pskId = pskId
        self._psk = psk
        self.coaptransport = coaptransport
        self.hostinfo = hostportjoin(host, None if port == COAPS_PORT else port)

        self._startup = asyncio.ensure_future(self._start())

    def _remove_from_pool(self):
        """Remove self from the MessageInterfaceTinyDTLS's pool, so that it
        will not be used in new requests.

        This is idempotent (to allow quick removal and still remove it in a
        finally clause) and not thread safe.
        """
        poolkey = (self._host, self._port, self._pskId)
        if self.coaptransport._pool.get(poolkey) is self:
            del self.coaptransport._pool[poolkey]

    def send(self, message):
        if self._queue is not None:
            self._queue.append(message)
        else:
            # most of the time that will have returned long ago
            self._retransmission_task.cancel()

            self._dtls_socket.write(self._connection, message)

            self._retransmission_task = asyncio.create_task(
                    self._run_retransmissions(),
                    **py38args(name="DTLS handshake retransmissions")
                    )

    log = property(lambda self: self.coaptransport.log)

    def _build_accessor(self, method, deadvalue):
        """Think self._build_accessor('_write')() == self._write(), just that
        it's returning a weak wrapper that allows refcounting-based GC to
        happen when the remote falls out of use"""
        weakself = weakref.ref(self)
        def wrapper(*args, __weakself=weakself, __method=method, __deadvalue=deadvalue):
            self = __weakself()
            if self is None:
                warnings.warn("DTLS module did not shut down the DTLSSocket "
                        "perfectly; it still tried to call %s in vain" %
                        __method)
                return __deadvalue
            return getattr(self, __method)(*args)
        wrapper.__name__ = "_build_accessor(%s)" % method
        return wrapper

    async def _start(self):
        from DTLSSocket import dtls

        self._dtls_socket = None

        self._connection = None

        try:
            self._transport, _ = await self.coaptransport.loop.create_datagram_endpoint(
                    self.SingleConnection.factory(self),
                    remote_addr=(self._host, self._port),
                    )

            self._dtls_socket = dtls.DTLS(
                    read=self._build_accessor("_read", 0),
                    write=self._build_accessor("_write", 0),
                    event=self._build_accessor("_event", 0),
                    pskId=self._pskId,
                    pskStore={self._pskId: self._psk},
                    )
            self._connection = self._dtls_socket.connect(_SENTINEL_ADDRESS, _SENTINEL_PORT)

            self._retransmission_task = asyncio.create_task(
                    self._run_retransmissions(),
                    **py38args(name="DTLS handshake retransmissions")
                    )

            self._connecting = asyncio.get_running_loop().create_future()
            await self._connecting

            queue = self._queue
            self._queue = None

            for message in queue:
                # could be a tad more efficient by stopping the retransmissions
                # in a go, then doing just the punch line and then starting it,
                # but practically this will be a single thing most of the time
                # anyway
                self.send(message)

            return

        except asyncio.CancelledError:
            # Can be removed starting with Python 3.8 as it's a workaround for
            # https://bugs.python.org/issue32528
            raise
        except Exception as e:
            self.coaptransport.ctx.dispatch_error(e, self)
        finally:
            if self._queue is None:
                # all worked, we're done here
                return

            self.shutdown()

    async def _run_retransmissions(self):
        while True:
            when = self._dtls_socket.checkRetransmit() / DTLS_TICKS_PER_SECOND
            if when == 0:
                return
            now = time.time() - DTLS_CLOCK_OFFSET
            await asyncio.sleep(when - now)


    def shutdown(self):
        self._remove_from_pool()

        self._startup.cancel()
        self._retransmission_task.cancel()

        if self._connection is not None:
            try:
                self._dtls_socket.close(self._connection)
            except Exception:
                pass # _dtls_socket actually does raise an empty Exception() here
        # doing this here allows the dtls socket to send a final word, but
        # by closing this, we protect the nascent next connection from any
        # delayed ICMP errors that might still wind up in the old socket
        self._transport.close()

    def __del__(self):
        # Breaking the loops between the DTLS object and this here to allow for
        # an orderly Alet (fatal, close notify) to go out -- and also because
        # DTLSSocket throws `TypeError: 'NoneType' object is not subscriptable`
        # from its destructor while the cyclical dependency is taken down.
        self.shutdown()

    def _inject_error(self, e):
        """Put an error to all pending operations on this remote, just as if it
        were raised inside the main loop."""

        self.coaptransport.ctx.dispatch_error(e, self)

        self.shutdown()

    # dtls callbacks

    def _read(self, sender, data):
        # ignoring sender: it's only _SENTINEL_*

        try:
            message = Message.decode(data, self)
        except error.UnparsableMessage:
            self.log.warning("Ignoring unparsable message from %s", sender)
            return len(data)

        self.coaptransport.ctx.dispatch_message(message)

        return len(data)

    def _write(self, recipient, data):
        # ignoring recipient: it's only _SENTINEL_*
        try:
            t = self._transport
        except Exception:
            # tinydtls sends callbacks very very late during shutdown (ie.
            # `hasattr` and `AttributeError` are all not available any more,
            # and even if the DTLSClientConnection class had a ._transport, it
            # would already be gone), and it seems even a __del__ doesn't help
            # break things up into the proper sequence.
            return 0
        t.sendto(data)
        return len(data)

    def _event(self, level, code):
        if (level, code) == (LEVEL_NOALERT, DTLS_EVENT_CONNECT):
            return
        elif (level, code) == (LEVEL_NOALERT, DTLS_EVENT_CONNECTED):
            self._connecting.set_result(True)
        elif (level, code) == (LEVEL_FATAL, CODE_CLOSE_NOTIFY):
            self._inject_error(CloseNotifyReceived())
        elif level == LEVEL_FATAL:
            self._inject_error(FatalDTLSError(code))
        else:
            self.log.warning("Unhandled alert level %d code %d", level, code)

    # transport protocol

    class SingleConnection:
        @classmethod
        def factory(cls, parent):
            return functools.partial(cls, weakref.ref(parent))

        def __init__(self, parent):
            self.parent = parent #: DTLSClientConnection

        def connection_made(self, transport):
            # only for for shutdown
            self.transport = transport

        def connection_lost(self, exc):
            pass

        def error_received(self, exc):
            parent = self.parent()
            if parent is None:
                self.transport.close()
                return
            parent._inject_error(exc)

        def datagram_received(self, data, addr):
            parent = self.parent()
            if parent is None:
                self.transport.close()
                return
            parent._dtls_socket.handleMessage(parent._connection, data)

class MessageInterfaceTinyDTLS(interfaces.MessageInterface):
    def __init__(self, ctx: interfaces.MessageManager, log, loop):
        self._pool = weakref.WeakValueDictionary({}) # see _connection_for_address

        self.ctx = ctx

        self.log = log
        self.loop = loop

    def _connection_for_address(self, host, port, pskId, psk):
        """Return a DTLSConnection to a given address. This will always give
        the same result for the same host/port combination, at least for as
        long as that result is kept alive (eg. by messages referring to it in
        their .remote) and while the connection has not failed."""

        try:
            return self._pool[(host, port, pskId)]
        except KeyError:
            self.log.info("No DTLS connection active to (%s, %s, %s), creating one", host, port, pskId)
            connection = DTLSClientConnection(host, port, pskId, psk, self)
            self._pool[(host, port, pskId)] = connection
            return connection

    @classmethod
    async def create_client_transport_endpoint(cls, ctx: interfaces.MessageManager, log, loop):
        return cls(ctx, log, loop)

    async def recognize_remote(self, remote):
        return isinstance(remote, DTLSClientConnection) and remote in self._pool.values()

    async def determine_remote(self, request):
        if request.requested_scheme != 'coaps':
            return None

        if request.unresolved_remote:
            host, port = hostportsplit(request.unresolved_remote)
            port = port or COAPS_PORT
        elif request.opt.uri_host:
            host = request.opt.uri_host
            port = request.opt.uri_port or COAPS_PORT
        else:
            raise ValueError("No location found to send message to (neither in .opt.uri_host nor in .remote)")

        dtlsparams = self.ctx.client_credentials.credentials_from_request(request)
        try:
            pskId, psk = dtlsparams.as_dtls_psk()
        except AttributeError:
            raise CredentialsMissingError("Credentials for requested URI are not compatible with DTLS-PSK")
        result = self._connection_for_address(host, port, pskId, psk)
        return result

    def send(self, message):
        message.remote.send(message.encode())

    async def shutdown(self):
        remaining_connections = list(self._pool.values())
        for c in remaining_connections:
            c.shutdown()



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import asyncio
import warnings

from ..message import UndecidedRemote
from .. import interfaces
from ..protocol import ClientObservation

from ..util import hostportsplit
from ..util.asyncio import py38args

class ProxyForwarder(interfaces.RequestProvider):
    """Object that behaves like a Context but only provides the request
    function and forwards all messages to a proxy.

    This is not a proxy itself, it is just the interface for an external
    one."""
    def __init__(self, proxy_address, context):
        if '://' not in proxy_address:
            warnings.warn("Proxy addresses without scheme are deprecated, "
                    "please specify like `coap://host`, `coap+tcp://ip:port` "
                    "etc.", DeprecationWarning)
            proxy_address = 'coap://' + proxy_address

        self.proxy_address = UndecidedRemote.from_pathless_uri(proxy_address)
        self.context = context

    proxy = property(lambda self: self._proxy)

    def request(self, message, **kwargs):
        if not isinstance(message.remote, UndecidedRemote):
            raise ValueError(
                "Message already has a configured "\
                "remote, set .opt.uri_{host,port} instead of remote")
        host, port = hostportsplit(message.remote.hostinfo)
        message.opt.uri_port = port
        message.opt.uri_host = host
        message.opt.proxy_scheme = self.proxy_address.scheme
        message.remote = self.proxy_address

        return self.context.request(message)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Container module, see submodules:

* :mod:`.client` -- using CoAP via a proxy server
* :mod:`.server` -- running a proxy server
"""



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Client components for registering with a resource-directory server"""

import asyncio
from urllib.parse import urljoin, urlparse
import logging

from socket import getfqdn

from ...util.linkformat import link_header

from ...message import Message
from ...numbers import GET, POST, DELETE, SERVICE_UNAVAILABLE, NOT_FOUND

__all__ = ['Registerer']

class Registerer:
    """Implementation of the client side of the registration of a resource
    directory. Until the object is :meth:`shut down <shutdown>`, it keeps the
    registration alive. It works both for registering the own context as well
    as for registering others (taking the role of a commissioning tool).

    The :attr:`state` attribute is kept up to date with an informal
    representation of whether the registration is currently active.

    If any step in the registration fails, the object will not retry
    indefinitely, and it will back off to earlier steps; after a limited number
    of retries after the last successful step, the object permanently enters a
    failed state. (In future extension, it might listen for external events
    that allow it to restart heuristics, like a new network interface coming
    up).

    The registration does not observe the resource list of the registered host
    (yet?), so registrations are only kept alive and never updated."""

    def __init__(self, context, rd=None, lt=90000, name_from_hostname=None,
            link_source=None, registration_parameters={},
            loggername='coap-rd-registerer'):
        """Use a ``context`` to create a registration at the Resource
        directiory at ``rd`` (defaulting to "find an RD yourself"; URIs should
        have no path component, unless the user wishes to sidestep the URI
        discovery step). It will be renewed every `lifetime` seconds.

        The registration data will be obtained by querying the context's site's
        ``.well-known/core`` resource, unless another source URI is given in
        ``link_source``, in which case this object acts as a Commissioning
        Tool.

        Parameters to pass with the registration can be given in
        ``registration_parameters``. ``lt`` and ``base`` default to the
        constructor arguments ``lt`` and ``link_source``, respectively.

        If ``name_from_hostname`` is True (or by default if ``ep`` is not
        present in ``registration_parameters``), the ``ep`` and ``d``
        registration parameters are derived from the host name."""

        self._context = context

        self._link_source = None
        self._link_data = None #: Message
        self._lt = lt
        self._initial_rd = rd

        self._directory_resource = None
        self._registration_resource = None

        self._registration_parameters = dict(registration_parameters)

        if name_from_hostname or (name_from_hostname is None and
                'ep' not in registration_parameters):
            ep, _, d = getfqdn().partition('.')
            self._registration_parameters['ep'] = ep
            if d:
                self._registration_parameters['d'] = d

        self.log = logging.getLogger(loggername)

        self._set_state("starting")
        self._task = asyncio.create_task(self._run())

    def __repr__(self):
        return "<%s at %#x: registering at %s as %s (currently %s)>" % (
                type(self).__name__, id(self), self._registration_resource or
                self._directory_resource or self._initial_rd,
                self._registration_parameters, self.state)

    def _set_state(self, newstate):
        self.log.debug("Entering state %s", newstate)
        self.state = newstate

    async def _fill_directory_resource(self, blacklist=set()):
        # FIXME: this should at some point catch network errors (short of
        # falling back to "RD discovery failed, backing off"), but that needs
        # falling back to other discovery methods here, and i don't know how
        # this will be done yet

        if self._directory_resource is not None:
            return

        if self._initial_rd is None:
            # FIXME: can't access DHCP options generically, dunno about SLAAC.
            # It seems to be a sane assumption that the best thing to do is to
            # assume we're on a big host and multicast is cheap here.
            self._directory_resource = await self._discovery_directory_uri('coap://[ff05::fd]]', blacklist=blacklist)
        else:
            components = urlparse(self._initial_rd)
            if not components.scheme or not components.netloc:
                raise ValueError("Explicit RD URIs need to contain scheme and path")

            if components.path:
                if self._initial_rd in blacklist:
                    raise self._UnrecoverableError("Explicitly configured RD was blacklisted")
                else:
                    self._directory_resource = self._initial_rd
            else:
                self._directory_resource = await self._discovery_directory_uri(self._initial_rd, blacklist=blacklist)

    async def _discovery_directory_uri(self, host, blacklist=set()):
        lookup_uri = urljoin(host,
                '/.well-known/core?rt=core.rd')

        try:
            # FIXME: this should be able to deal with multicasts
            response = await self._context.request(
                    Message(code=GET, uri=lookup_uri, accept=40)
                ).response_raising
            links = link_header.parse(response.payload.decode('utf8'))
        except (UnicodeDecodeError, link_header.ParseException):
            self.log.error("Error parsing the RD's self description")
            raise

        addresses = [l.get_target(response.get_request_uri())
                for l in links.links
                if 'core.rd' in " ".join(l.rt).split(" ")]
        unfiltered_addresses = len(addresses)
        addresses = [a for a in addresses if a not in blacklist]
        if not addresses:
            if len(addresses) != unfiltered_addresses:
                raise self._UnrecoverableError("All discovered Directory Resources are blacklisted")
            else:
                raise self._UnrecoverableError("No registration interface found in RD's response")

        if len(addresses) > 1:
            self.log.warn("More than one registration interface found,"
                    " picking the first")

        return addresses[0]

    async def _obtain_link_data(self):
        """Store a message describing the data to be POSTed to the
        registration interface.

        This needs to be in :class:`Message` format, but doesn't need to have
        any particular code set yet (that gets set later anyway), so in effect,
        the response message from the con can be returned as is.
        """

        if self._link_source is None:
            self._link_data = Message(
                    content_format=40,
                    payload=str(self._context.serversite.get_resources_as_linkheader()).encode('utf8')
                    )

        else:
            self._link_data = await self._context.request(
                    Message(code=GET, uri=urljoin(self._link_source, '/.well-known/core'))
                    ).response_raising

    class _RetryableError(RuntimeError):
        """Raised when an initial registration or update rails in a way that
        warrants rediscovery of the RD"""

    class _UnrecoverableError(RuntimeError):
        """Raised when the RD registration process runs out of options
        (typically with a descriptive message)"""

    async def _request_with_retries(self, message):
        # FIXME: response_nonraising gives 5.00 now, but for debugging we might
        # want to show something better, and for URI discovery, we should not
        # consider this a final error
        response = await self._context.request(message).response_nonraising

        unavailable_retries = 0
        while response.code == SERVICE_UNAVAILABLE and response.opt.max_age is not None:
            if unavailable_retries > 6:
                raise self._RetryableError("RD responded with Service Unavailable too often")
            self.log.info("RD asked to retry the operation later")
            await asyncio.sleep(max(response.opt.max_age, 2**(unavailable_retries)))
            response = await self._context.request(message).response_nonraising

        return response

    async def _register(self):
        initial_message = self._link_data.copy(code=POST, uri=self._directory_resource)
        base_query = {}
        if self._lt != 90000:
            base_query['lt'] = str(self._lt)
        if self._link_source is not None:
            base_query['base'] = self._link_source
        query = dict(base_query, **self._registration_parameters)

        initial_message.opt.uri_query = initial_message.opt.uri_query + \
                tuple("%s=%s" % (k, v) for (k, v) in query.items())

        response = await self._request_with_retries(initial_message)

        if not response.code.is_successful():
            raise self._RetryableError("RD responded with odd error: %s / %r" % (response.code, response.payload))

        if not response.opt.location_path:
            raise self._RetryableError("RD responded without a location")

        # FIXME this should probably be available from the API, and consider location_query etc
        self._registration_resource = urljoin(response.get_request_uri(), "/" + "/".join(response.opt.location_path))

    async def _renew_registration(self):
        update_message = Message(code=POST, uri=self._registration_resource)

        response = await self._request_with_retries(update_message)

        if response.code == NOT_FOUND:
            raise self._RetryableError("RD forgot about the registration")

        if not response.code.is_successful():
            raise self._RetryableError("RD responded with odd error: %s / %r" % (response.code, response.payload))

    async def _run(self):
        obtain = asyncio.create_task(self._obtain_link_data())

        try:
            await self._run_inner(obtain)
        except asyncio.CancelledError:
            self._set_state('cancelled')
            pass
        except self._UnrecoverableError as e:
            self._set_state('failed')
            self.log.error("Aborting RD discovery: %s", e.args[0])
        except Exception as e:
            self._set_state('failed')
            self.log.error("An error occurred during RD registration, not pursuing registration any further:", exc_info=e)
        finally:
            obtain.cancel()

    async def _run_inner(self, obtain):
        errors = 0
        errors_max = 5
        failed_initialization = set()
        try_reuse_discovery = False

        while True:
            if try_reuse_discovery:
                try_reuse_discovery = False
            else:
                self._set_state('discovering')

                for i in range(4):
                    if i:
                        self.log.info("Waiting to retry RD discovery")
                        await asyncio.sleep(2 * 3 ** (i - 1)) # arbitrary fall-off
                    await self._fill_directory_resource(blacklist=failed_initialization)
                    break
                else:
                    self.log.error("Giving up RD discovery")
                    break

            await obtain

            self._set_state("registering")

            try:
                await self._register()
            except self._RetryableError as e:
                errors += 1
                if errors < errors_max:
                    self.log.warning("Initial registration failed (%s), blacklisting RD URI and retrying discovery", e)
                    failed_initialization.add(self._directory_resource)
                    self._directory_resource = None
                    continue
                else:
                    self.log.error("Giving up after too many failed initial registrations")
                    break

            # registration is active, keep it that way.

            # things look good enough to forget about past bad experiences.
            # could move this to the end of the following loop if worries come
            # up of having picked a bad RD that supports registration but not
            # registration updates
            errors = 0
            failed_initialization = set()
            try_reuse_discovery = True

            while True:
                self._set_state("registered")

                # renew 60 seconds before timeout, unless that's before the 75% mark (then wait for that)
                await asyncio.sleep(self._lt - 60 if self._lt > 240 else self._lt * 3 // 4)

                self._set_state("renewing")

                try:
                    await self._renew_registration()
                except self._RetryableError as e:
                    self.log.warning("Registration update failed (%s), retrying with new registration", e)
                    break

    async def shutdown(self):
        """Delete the registration. This will not raise any resulting error
        messages but just log them, same as any errors occurring during the
        registration will only be logged."""
        self._task.cancel()

        if self._registration_resource is None:
            return

        try:
            await self._context.request(
                    Message(code=DELETE, uri=self._registration_resource)
                    ).response_raising
        except Exception as e:
            self.log.error("Error deregistering from the RD", exc_info=e)
